# This code is part of Ansible, but is an independent component.
# This particular file snippet, and this file snippet only, is BSD licensed.
# Modules you write using this snippet, which is embedded dynamically by Ansible
# still belong to the author of the module, and may assign their own license
# to the complete work.
#
# (c) 2016 Red Hat Inc.
#
# Simplified BSD License (see LICENSES/BSD-2-Clause.txt or https://opensource.org/licenses/BSD-2-Clause)
# SPDX-License-Identifier: BSD-2-Clause

from __future__ import absolute_import, division, print_function


__metaclass__ = type

# Networking tools for network modules only

import ast
import json
import operator
import re
import socket

from copy import deepcopy
from functools import reduce  # forward compatibility for Python 3
from io import StringIO
from itertools import chain

from ansible.module_utils import basic
from ansible.module_utils._text import to_bytes, to_text
from ansible.module_utils.common._collections_compat import Mapping
from ansible.module_utils.parsing.convert_bool import boolean
from ansible.module_utils.six import iteritems, string_types


try:
    from jinja2 import Environment, StrictUndefined
    from jinja2.exceptions import UndefinedError

    HAS_JINJA2 = True
except ImportError:
    HAS_JINJA2 = False

try:
    import yaml

    try:
        # use C version if possible for speedup
        from yaml import CSafeLoader as SafeLoader
    except ImportError:
        from yaml import SafeLoader
    HAS_YAML = True
except ImportError:
    HAS_YAML = False

OPERATORS = frozenset(["ge", "gt", "eq", "neq", "lt", "le"])
ALIASES = frozenset([("min", "ge"), ("max", "le"), ("exactly", "eq"), ("neq", "ne")])

OPTION_METADATA = (
    "type",
    "choices",
    "default",
    "required",
    "aliases",
    "elements",
    "fallback",
    "no_log",
    "apply_defaults",
    "deprecated_aliases",
    "removed_in_version",
)
OPTION_CONDITIONALS = (
    "mutually_exclusive",
    "required_one_of",
    "required_together",
    "required_by",
    "required_if",
)

VALID_ANSIBLEMODULE_ARGS = (
    "bypass_checks",
    "no_log",
    "add_file_common_args",
    "supports_check_mode",
) + OPTION_CONDITIONALS


def to_list(val):
    if isinstance(val, (list, tuple, set)):
        return list(val)
    elif val is not None:
        return [val]
    else:
        return list()


def to_lines(stdout):
    for item in stdout:
        if isinstance(item, string_types):
            item = to_text(item).split("\n")
        yield item


def transform_commands(module):
    transform = EntityCollection(
        module,
        dict(
            command=dict(key=True),
            output=dict(),
            prompt=dict(type="list"),
            answer=dict(type="list"),
            newline=dict(type="bool", default=True),
            sendonly=dict(type="bool", default=False),
            check_all=dict(type="bool", default=False),
        ),
    )

    return transform(module.params["commands"])


def sort_list(val):
    if isinstance(val, list):
        if isinstance(val[0], dict):
            sorted_keys = [tuple(sorted(dict_.keys())) for dict_ in val]
            # All keys should be identical
            if len(set(sorted_keys)) != 1:
                raise ValueError("dictionaries do not match")

            return sorted(val, key=lambda d: tuple(d[k] for k in sorted_keys[0]))
        return sorted(val)
    return val


class Entity(object):
    """Transforms a dict to with an argument spec

    This class will take a dict and apply an Ansible argument spec to the
    values.  The resulting dict will contain all of the keys in the param
    with appropriate values set.

    Example::

        argument_spec = dict(
            command=dict(key=True),
            display=dict(default='text', choices=['text', 'json']),
            validate=dict(type='bool')
        )
        transform = Entity(module, argument_spec)
        value = dict(command='foo')
        result = transform(value)
        print(result)
        {'command': 'foo', 'display': 'text', 'validate': None}

    Supported argument spec:
        * key - specifies how to map a single value to a dict
        * read_from - read and apply the argument_spec from the module
        * required - a value is required
        * type - type of value (uses AnsibleModule type checker)
        * fallback - implements fallback function
        * choices - set of valid options
        * default - default value
    """

    def __init__(self, module, attrs=None, args=None, keys=None, from_argspec=False):
        args = [] if args is None else args

        self._attributes = attrs or {}
        self._module = module

        for arg in args:
            self._attributes[arg] = dict()
            if from_argspec:
                self._attributes[arg]["read_from"] = arg
            if keys and arg in keys:
                self._attributes[arg]["key"] = True

        self.attr_names = frozenset(self._attributes.keys())

        _has_key = False

        for name, attr in iteritems(self._attributes):
            if attr.get("read_from"):
                if attr["read_from"] not in self._module.argument_spec:
                    module.fail_json(msg="argument %s does not exist" % attr["read_from"])
                spec = self._module.argument_spec.get(attr["read_from"])
                for key, value in iteritems(spec):
                    if key not in attr:
                        attr[key] = value

            if attr.get("key"):
                if _has_key:
                    module.fail_json(msg="only one key value can be specified")
                _has_key = True
                attr["required"] = True

    def serialize(self):
        return self._attributes

    def to_dict(self, value):
        obj = {}
        for name, attr in iteritems(self._attributes):
            if attr.get("key"):
                obj[name] = value
            else:
                obj[name] = attr.get("default")
        return obj

    def __call__(self, value, strict=True):
        if not isinstance(value, dict):
            value = self.to_dict(value)

        if strict:
            unknown = set(value).difference(self.attr_names)
            if unknown:
                self._module.fail_json(msg="invalid keys: %s" % ",".join(unknown))

        for name, attr in iteritems(self._attributes):
            if value.get(name) is None:
                value[name] = attr.get("default")

            if attr.get("fallback") and not value.get(name):
                fallback = attr.get("fallback", (None,))
                fallback_strategy = fallback[0]
                fallback_args = []
                fallback_kwargs = {}
                if fallback_strategy is not None:
                    for item in fallback[1:]:
                        if isinstance(item, dict):
                            fallback_kwargs = item
                        else:
                            fallback_args = item
                    try:
                        value[name] = fallback_strategy(*fallback_args, **fallback_kwargs)
                    except basic.AnsibleFallbackNotFound:
                        continue

            if attr.get("required") and value.get(name) is None:
                self._module.fail_json(msg="missing required attribute %s" % name)

            if "choices" in attr:
                if value[name] not in attr["choices"]:
                    self._module.fail_json(
                        msg="%s must be one of %s, got %s"
                        % (name, ", ".join(attr["choices"]), value[name])
                    )

            if value[name] is not None:
                value_type = attr.get("type", "str")
                type_checker = self._module._CHECK_ARGUMENT_TYPES_DISPATCHER[value_type]
                type_checker(value[name])
            elif value.get(name):
                value[name] = self._module.params[name]

        return value


class EntityCollection(Entity):
    """Extends ```Entity``` to handle a list of dicts"""

    def __call__(self, iterable, strict=True):
        if iterable is None:
            iterable = [super(EntityCollection, self).__call__(self._module.params, strict)]

        if not isinstance(iterable, (list, tuple)):
            self._module.fail_json(msg="value must be an iterable")

        return [(super(EntityCollection, self).__call__(i, strict)) for i in iterable]


class ComplexList(EntityCollection):
    """Alternate name for EntityCollection for backwards compatibility"""

    def __init__(self, attrs, module, *args, **kwargs):
        super(ComplexList, self).__init__(module, attrs, *args, **kwargs)


def dict_diff(base, comparable):
    """Generate a dict object of differences

    This function will compare two dict objects and return the difference
    between them as a dict object.  For scalar values, the key will reflect
    the updated value.  If the key does not exist in `comparable`, then then no
    key will be returned.  For lists, the value in comparable will wholly replace
    the value in base for the key.  For dicts, the returned value will only
    return keys that are different.

    :param base: dict object to base the diff on
    :param comparable: dict object to compare against base

    :returns: new dict object with differences
    """
    if not isinstance(base, dict):
        raise AssertionError("`base` must be of type <dict>")
    if not isinstance(comparable, dict):
        if comparable is None:
            comparable = dict()
        else:
            raise AssertionError("`comparable` must be of type <dict>")

    updates = dict()

    for key, value in iteritems(base):
        if isinstance(value, dict):
            item = comparable.get(key)
            if item is not None:
                sub_diff = dict_diff(value, comparable[key])
                if sub_diff:
                    updates[key] = sub_diff
        else:
            comparable_value = comparable.get(key)
            if comparable_value is not None:
                if sort_list(base[key]) != sort_list(comparable_value):
                    updates[key] = comparable_value

    for key in set(comparable.keys()).difference(base.keys()):
        updates[key] = comparable.get(key)

    return updates


def dict_merge(base, other):
    """Return a new dict object that combines base and other

    This will create a new dict object that is a combination of the key/value
    pairs from base and other.  When both keys exist, the value will be
    selected from other.  If the value is a list object, the two lists will
    be combined and duplicate entries removed.

    :param base: dict object to serve as base
    :param other: dict object to combine with base

    :returns: new combined dict object
    """
    if not isinstance(base, dict):
        raise AssertionError("`base` must be of type <dict>")
    if not isinstance(other, dict):
        raise AssertionError("`other` must be of type <dict>")

    combined = dict()

    for key, value in iteritems(deepcopy(base)):
        if isinstance(value, dict):
            if key in other:
                item = other.get(key)
                if item is not None:
                    if isinstance(other[key], Mapping):
                        combined[key] = dict_merge(value, other[key])
                    else:
                        combined[key] = other[key]
                else:
                    combined[key] = item
            else:
                combined[key] = value
        elif isinstance(value, list):
            if key in other:
                item = other.get(key)
                if item is not None:
                    try:
                        combined[key] = list(set(chain(value, item)))
                    except TypeError:
                        value.extend([i for i in item if i not in value])
                        combined[key] = value
                else:
                    combined[key] = item
            else:
                combined[key] = value
        else:
            if key in other:
                other_value = other.get(key)
                if other_value is not None:
                    if sort_list(base[key]) != sort_list(other_value):
                        combined[key] = other_value
                    else:
                        combined[key] = value
                else:
                    combined[key] = other_value
            else:
                combined[key] = value

    for key in set(other.keys()).difference(base.keys()):
        combined[key] = other.get(key)

    return combined


def param_list_to_dict(param_list, unique_key="name", remove_key=True):
    """Rotates a list of dictionaries to be a dictionary of dictionaries.

    :param param_list: The aforementioned list of dictionaries
    :param unique_key: The name of a key which is present and unique in all of param_list's dictionaries. The value
    behind this key will be the key each dictionary can be found at in the new root dictionary
    :param remove_key: If True, remove unique_key from the individual dictionaries before returning.
    """
    param_dict = {}
    for params in param_list:
        params = params.copy()
        if remove_key:
            name = params.pop(unique_key)
        else:
            name = params.get(unique_key)
        param_dict[name] = params

    return param_dict


def conditional(expr, val, cast=None):
    match = re.match(r"^(.+)\((.+)\)$", str(expr), re.I)
    if match:
        op, arg = match.groups()
    else:
        op = "eq"
        if " " in str(expr):
            raise AssertionError("invalid expression: cannot contain spaces")
        arg = expr

    if cast is None and val is not None:
        arg = type(val)(arg)
    elif callable(cast):
        arg = cast(arg)
        val = cast(val)

    op = next((oper for alias, oper in ALIASES if op == alias), op)

    if not hasattr(operator, op) and op not in OPERATORS:
        raise ValueError("unknown operator: %s" % op)

    func = getattr(operator, op)
    return func(val, arg)


def ternary(value, true_val, false_val):
    """value ? true_val : false_val"""
    if value:
        return true_val
    else:
        return false_val


def remove_default_spec(spec):
    for item in spec:
        if "default" in spec[item]:
            del spec[item]["default"]


def validate_ip_address(address):
    try:
        socket.inet_aton(address)
    except socket.error:
        return False
    return address.count(".") == 3


def validate_ip_v6_address(address):
    try:
        socket.inet_pton(socket.AF_INET6, address)
    except socket.error:
        return False
    return True


def validate_prefix(prefix):
    if prefix and not 0 <= int(prefix) <= 32:
        return False
    return True


def load_provider(spec, args):
    provider = args.get("provider") or {}
    for key, value in iteritems(spec):
        if key not in provider:
            try:
                # Get fallback if defined, and valid
                provider[key] = _fallback(value["fallback"])
            except (basic.AnsibleFallbackNotFound, KeyError):
                # Get default if defined, otherwise set to None
                provider[key] = value.get("default")
    if "authorize" in provider:
        # Coerce authorize to provider if a string has somehow snuck in.
        provider["authorize"] = boolean(provider["authorize"] or False)
    args["provider"] = provider
    return provider


def _fallback(fallback):
    strategy = fallback[0]
    args = []
    kwargs = {}

    for item in fallback[1:]:
        if isinstance(item, dict):
            kwargs = item
        else:
            args = item

    return strategy(*args, **kwargs)


def generate_dict(spec):
    """
    Generate dictionary which is in sync with argspec

    :param spec: A dictionary that is the argspec of the module
    :rtype: A dictionary
    :returns: A dictionary in sync with argspec with default value
    """
    obj = {}
    if not spec:
        return obj

    for key, val in iteritems(spec):
        if "default" in val:
            dct = {key: val["default"]}
        elif "type" in val and val["type"] == "dict":
            dct = {key: generate_dict(val["options"])}
        else:
            dct = {key: None}
        obj.update(dct)
    return obj


def parse_conf_arg(cfg, arg):
    """
    Parse config based on argument

    :param cfg: A text string which is a line of configuration.
    :param arg: A text string which is to be matched.
    :rtype: A text string
    :returns: A text string if match is found
    """
    match = re.search(r"%s (.+)(\n|$)" % arg, cfg, re.M)
    if match:
        result = match.group(1).strip()
    else:
        result = None
    return result


def parse_conf_cmd_arg(cfg, cmd, res1, res2=None, delete_str="no"):
    """
    Parse config based on command

    :param cfg: A text string which is a line of configuration.
    :param cmd: A text string which is the command to be matched
    :param res1: A text string to be returned if the command is present
    :param res2: A text string to be returned if the negate command
                 is present
    :param delete_str: A text string to identify the start of the
                 negate command
    :rtype: A text string
    :returns: A text string if match is found
    """
    match = re.search(r"\n\s+%s(\n|$)" % cmd, cfg)
    if match:
        return res1
    if res2 is not None:
        match = re.search(r"\n\s+%s %s(\n|$)" % (delete_str, cmd), cfg)
        if match:
            return res2
    return None


def get_xml_conf_arg(cfg, path, data="text"):
    """
    :param cfg: The top level configuration lxml Element tree object
    :param path: The relative xpath w.r.t to top level element (cfg)
           to be searched in the xml hierarchy
    :param data: The type of data to be returned for the matched xml node.
        Valid values are text, tag, attrib, with default as text.
    :return: Returns the required type for the matched xml node or else None
    """
    match = cfg.xpath(path)
    if len(match):
        if data == "tag":
            result = getattr(match[0], "tag")
        elif data == "attrib":
            result = getattr(match[0], "attrib")
        else:
            result = getattr(match[0], "text")
    else:
        result = None
    return result


def remove_empties(cfg_dict):
    """
    Generate final config dictionary

    :param cfg_dict: A dictionary parsed in the facts system
    :rtype: A dictionary
    :returns: A dictionary by eliminating keys that have null values
    """
    final_cfg = {}
    if not cfg_dict:
        return final_cfg

    for key, val in iteritems(cfg_dict):
        dct = None
        if isinstance(val, dict):
            child_val = remove_empties(val)
            if child_val:
                dct = {key: child_val}
        elif isinstance(val, list) and val and all(isinstance(x, dict) for x in val):
            child_val = [remove_empties(x) for x in val]
            if child_val:
                dct = {key: child_val}
        elif val not in [None, [], {}, (), ""]:
            dct = {key: val}
        if dct:
            final_cfg.update(dct)
    return final_cfg


def validate_config(spec, data):
    """
    Validate if the input data against the AnsibleModule spec format
    :param spec: Ansible argument spec
    :param data: Data to be validated
    :return:
    """
    params = basic._ANSIBLE_ARGS
    basic._ANSIBLE_ARGS = to_bytes(json.dumps({"ANSIBLE_MODULE_ARGS": data}))
    validated_data = basic.AnsibleModule(spec).params
    basic._ANSIBLE_ARGS = params
    return validated_data


def search_obj_in_list(name, lst, key="name"):
    if not lst:
        return None
    else:
        for item in lst:
            if item.get(key) == name:
                return item


def get_from_dict(data_dict, keypath):
    """get from dictionary"""
    map_list = keypath.split(".")
    try:
        return reduce(operator.getitem, map_list, data_dict)
    except KeyError:
        return None


def compare_partial_dict(want, have, compare_keys):
    """compare"""
    rmkeys = [ckey[1:] for ckey in compare_keys if ckey.startswith("!")]
    kkeys = [ckey for ckey in compare_keys if not ckey.startswith("!")]

    wantd = {}
    for key, val in want.items():
        if key not in rmkeys or key in kkeys:
            wantd[key] = val

    haved = {}
    for key, val in have.items():
        if key not in rmkeys or key in kkeys:
            haved[key] = val

    return wantd == haved


class Template:
    def __init__(self):
        if not HAS_JINJA2:
            raise ImportError(
                "jinja2 is required but does not appear to be installed.  "
                "It can be installed using `pip install jinja2`"
            )

        self.env = Environment(undefined=StrictUndefined)
        self.env.filters.update({"ternary": ternary, "all": all, "any": any})

    def __call__(self, value, variables=None, fail_on_undefined=True):
        variables = variables or {}

        if not self.contains_vars(value):
            return value

        try:
            value = self.env.from_string(value).render(variables)
        except UndefinedError:
            if not fail_on_undefined:
                return None
            raise

        if value:
            try:
                return ast.literal_eval(value)
            except Exception:
                return str(value)
        else:
            return None

    def contains_vars(self, data):
        if isinstance(data, string_types):
            for marker in (
                self.env.block_start_string,
                self.env.variable_start_string,
                self.env.comment_start_string,
            ):
                if marker in data:
                    return True
        return False


def extract_argspec(doc_obj, argpsec):
    options_obj = doc_obj.get("options")
    for okey, ovalue in iteritems(options_obj):
        argpsec[okey] = {}
        for metakey in list(ovalue):
            if metakey == "suboptions":
                argpsec[okey].update({"options": {}})
                suboptions_obj = {"options": ovalue["suboptions"]}
                extract_argspec(suboptions_obj, argpsec[okey]["options"])
            elif metakey in OPTION_METADATA + OPTION_CONDITIONALS:
                argpsec[okey].update({metakey: ovalue[metakey]})


# TODO: Support extends_documentation_fragment
def convert_doc_to_ansible_module_kwargs(doc):
    doc_obj = yaml.load(StringIO(doc), SafeLoader)
    argspec = {}
    spec = {}
    extract_argspec(doc_obj, argspec)
    spec.update({"argument_spec": argspec})
    for item in doc_obj:
        if item in VALID_ANSIBLEMODULE_ARGS:
            spec.update({item: doc_obj[item]})
    return spec



# (c) 2016 Red Hat Inc.
# (c) 2017 Ansible Project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
author:
 - Ansible Networking Team (@ansible-network)
name: netconf
short_description: Provides a persistent connection using the netconf protocol
description:
- This connection plugin provides a connection to remote devices over the SSH NETCONF
  subsystem.  This connection plugin is typically used by network devices for sending
  and receiving RPC calls over NETCONF.
- Note this connection plugin requires ncclient to be installed on the local Ansible
  controller.
version_added: 1.0.0
requirements:
- ncclient
extends_documentation_fragment:
- ansible.netcommon.connection_persistent
options:
  host:
    description:
    - Specifies the remote device FQDN or IP address to establish the SSH connection
      to.
    default: inventory_hostname
    type: string
    vars:
    - name: inventory_hostname
    - name: ansible_host
  port:
    type: int
    description:
    - Specifies the port on the remote device that listens for connections when establishing
      the SSH connection.
    default: 830
    ini:
    - section: defaults
      key: remote_port
    env:
    - name: ANSIBLE_REMOTE_PORT
    vars:
    - name: ansible_port
  network_os:
    description:
    - Configures the device platform network operating system.  This value is used
      to load a device specific netconf plugin.  If this option is not configured
      (or set to C(auto)), then Ansible will attempt to guess the correct network_os
      to use. If it can not guess a network_os correctly it will use C(default).
    type: string
    vars:
    - name: ansible_network_os
  remote_user:
    description:
    - The username used to authenticate to the remote device when the SSH connection
      is first established.  If the remote_user is not specified, the connection will
      use the username of the logged in user.
    - Can be configured from the CLI via the C(--user) or C(-u) options.
    type: string
    ini:
    - section: defaults
      key: remote_user
    env:
    - name: ANSIBLE_REMOTE_USER
    vars:
    - name: ansible_user
  password:
    description:
    - Configures the user password used to authenticate to the remote device when
      first establishing the SSH connection.
    type: string
    vars:
    - name: ansible_password
    - name: ansible_ssh_pass
    - name: ansible_ssh_password
    - name: ansible_netconf_password
  private_key_file:
    description:
    - The private SSH key or certificate file used to authenticate to the remote device
      when first establishing the SSH connection.
    type: string
    ini:
    - section: defaults
      key: private_key_file
    env:
    - name: ANSIBLE_PRIVATE_KEY_FILE
    vars:
    - name: ansible_private_key_file
  look_for_keys:
    default: true
    description:
    - Enables looking for ssh keys in the usual locations for ssh keys (e.g. :file:`~/.ssh/id_*`).
    env:
    - name: ANSIBLE_PARAMIKO_LOOK_FOR_KEYS
    ini:
    - section: paramiko_connection
      key: look_for_keys
    type: boolean
  host_key_checking:
    description: Set this to "False" if you want to avoid host key checking by the
      underlying tools Ansible uses to connect to the host
    type: boolean
    default: true
    env:
    - name: ANSIBLE_HOST_KEY_CHECKING
    - name: ANSIBLE_SSH_HOST_KEY_CHECKING
    - name: ANSIBLE_NETCONF_HOST_KEY_CHECKING
    ini:
    - section: defaults
      key: host_key_checking
    - section: paramiko_connection
      key: host_key_checking
    vars:
    - name: ansible_host_key_checking
    - name: ansible_ssh_host_key_checking
    - name: ansible_netconf_host_key_checking
  proxy_command:
    default: ''
    description:
      - Proxy information for running the connection via a jumphost.
      - This requires ncclient >= 0.6.10 to be installed on the controller.
    type: string
    env:
      - name: ANSIBLE_NETCONF_PROXY_COMMAND
    ini:
      - {key: proxy_command, section: paramiko_connection}
    vars:
      - name: ansible_paramiko_proxy_command
      - name: ansible_netconf_proxy_command
  netconf_ssh_config:
    description:
    - This variable is used to enable bastion/jump host with netconf connection. If
      set to True the bastion/jump host ssh settings should be present in ~/.ssh/config
      file, alternatively it can be set to custom ssh configuration file path to read
      the bastion/jump host settings.
    type: string
    ini:
    - section: netconf_connection
      key: ssh_config
    env:
    - name: ANSIBLE_NETCONF_SSH_CONFIG
    vars:
    - name: ansible_netconf_ssh_config
"""

import json
import logging
import os

from ansible.errors import AnsibleConnectionFailure, AnsibleError
from ansible.module_utils._text import to_bytes, to_native, to_text
from ansible.module_utils.basic import missing_required_lib
from ansible.module_utils.parsing.convert_bool import BOOLEANS_FALSE, BOOLEANS_TRUE
from ansible.module_utils.six import PY3
from ansible.module_utils.six.moves import cPickle
from ansible.playbook.play_context import PlayContext
from ansible.plugins.connection import ensure_connect
from ansible.plugins.loader import netconf_loader

from ansible_collections.ansible.netcommon.plugins.plugin_utils.connection_base import (
    NetworkConnectionBase,
)
from ansible_collections.ansible.netcommon.plugins.plugin_utils.version import Version


try:
    from ncclient import __version__ as NCCLIENT_VERSION
    from ncclient import manager
    from ncclient.operations import RPCError
    from ncclient.transport.errors import AuthenticationError, SSHUnknownHostError
    from ncclient.xml_ import to_ele, to_xml
    from paramiko import ProxyCommand

    HAS_NCCLIENT = True
    NCCLIENT_IMP_ERR = None
# paramiko and gssapi are incompatible and raise AttributeError not ImportError
# When running in FIPS mode, cryptography raises InternalError
# https://bugzilla.redhat.com/show_bug.cgi?id=1778939
except Exception as err:
    HAS_NCCLIENT = False
    NCCLIENT_IMP_ERR = err

logging.getLogger("ncclient").setLevel(logging.INFO)


class Connection(NetworkConnectionBase):
    """NetConf connections"""

    transport = "ansible.netcommon.netconf"
    has_pipelining = False

    def __init__(self, play_context, new_stdin, *args, **kwargs):
        super(Connection, self).__init__(play_context, new_stdin, *args, **kwargs)

        # If network_os is not specified then set the network os to auto
        # This will be used to trigger the use of guess_network_os when connecting.
        self._network_os = self._network_os or "auto"

        self.netconf = netconf_loader.get(self._network_os, self)
        if self.netconf:
            self._sub_plugin = {
                "type": "netconf",
                "name": self.netconf._load_name,
                "obj": self.netconf,
            }
            self.queue_message(
                "vvvv",
                "loaded netconf plugin %s from path %s for network_os %s"
                % (
                    self.netconf._load_name,
                    self.netconf._original_path,
                    self._network_os,
                ),
            )
        else:
            self.netconf = netconf_loader.get("default", self)
            self._sub_plugin = {
                "type": "netconf",
                "name": "default",
                "obj": self.netconf,
            }
            self.queue_message(
                "vvvv",
                "unable to load netconf plugin for network_os %s, falling back to default plugin"
                % self._network_os,
            )

        self.queue_message("log", "network_os is set to %s" % self._network_os)
        self._manager = None
        self.key_filename = None
        self._ssh_config = None

    def exec_command(self, cmd, in_data=None, sudoable=True):
        """Sends the request to the node and returns the reply
        The method accepts two forms of request.  The first form is as a byte
        string that represents xml string be send over netconf session.
        The second form is a json-rpc (2.0) byte string.
        """
        if self._manager:
            # to_ele operates on native strings
            request = to_ele(to_native(cmd, errors="surrogate_or_strict"))

            if request is None:
                return "unable to parse request"

            try:
                reply = self._manager.rpc(request)
            except RPCError as exc:
                error = self.internal_error(
                    data=to_text(to_xml(exc.xml), errors="surrogate_or_strict")
                )
                return json.dumps(error)

            return reply.data_xml
        else:
            return super(Connection, self).exec_command(cmd, in_data, sudoable)

    def update_play_context(self, pc_data):
        """Updates the play context information for the connection"""
        pc_data = to_bytes(pc_data)
        if PY3:
            pc_data = cPickle.loads(pc_data, encoding="bytes")
        else:
            pc_data = cPickle.loads(pc_data)
        play_context = PlayContext()
        play_context.deserialize(pc_data)
        self._play_context = play_context

    @property
    @ensure_connect
    def manager(self):
        return self._manager

    def _get_proxy_command(self, port=22):
        proxy_command = None

        # TO-DO: Add logic to scan ssh_* args to read ProxyCommand

        proxy_command = self.get_option("proxy_command")

        sock = None
        if proxy_command:
            if Version(NCCLIENT_VERSION) < "0.6.10":
                raise AnsibleError(
                    "Configuring jumphost settings through ProxyCommand is unsupported in ncclient version %s. "
                    "Please upgrade to ncclient 0.6.10 or newer." % NCCLIENT_VERSION
                )

            replacers = {
                "%h": self._play_context.remote_addr,
                "%p": port,
                "%r": self._play_context.remote_user,
            }

            for find, replace in replacers.items():
                proxy_command = proxy_command.replace(find, str(replace))
            sock = ProxyCommand(proxy_command)

        return sock

    def _connect(self):
        if not HAS_NCCLIENT:
            raise AnsibleError(
                "%s: %s"
                % (
                    missing_required_lib("ncclient"),
                    to_native(NCCLIENT_IMP_ERR),
                )
            )

        self.queue_message("log", "ssh connection done, starting ncclient")

        allow_agent = True
        if self._play_context.password is not None:
            allow_agent = False
        setattr(self._play_context, "allow_agent", allow_agent)

        self.key_filename = self._play_context.private_key_file or self.get_option(
            "private_key_file"
        )
        if self.key_filename:
            self.key_filename = str(os.path.expanduser(self.key_filename))

        self._ssh_config = self.get_option("netconf_ssh_config")
        if self._ssh_config in BOOLEANS_TRUE:
            self._ssh_config = True
        elif self._ssh_config in BOOLEANS_FALSE:
            self._ssh_config = None

        # Try to guess the network_os if the network_os is set to auto
        if self._network_os == "auto":
            for cls in netconf_loader.all(class_only=True):
                network_os = cls.guess_network_os(self)
                if network_os:
                    self.queue_message("vvv", "discovered network_os %s" % network_os)
                    self._network_os = network_os

        # If we have tried to detect the network_os but were unable to i.e. network_os is still 'auto'
        # then use default as the network_os

        if self._network_os == "auto":
            # Network os not discovered. Set it to default
            self.queue_message(
                "vvv",
                "Unable to discover network_os. Falling back to default.",
            )
            self._network_os = "default"
        try:
            ncclient_device_handler = self.netconf.get_option("ncclient_device_handler")
        except KeyError:
            ncclient_device_handler = "default"
        self.queue_message(
            "vvv",
            "identified ncclient device handler: %s." % ncclient_device_handler,
        )
        device_params = {"name": ncclient_device_handler}

        try:
            port = self._play_context.port or 830
            self.queue_message(
                "vvv",
                "ESTABLISH NETCONF SSH CONNECTION FOR USER: %s on PORT %s TO %s WITH SSH_CONFIG = %s"
                % (
                    self._play_context.remote_user,
                    port,
                    self._play_context.remote_addr,
                    self._ssh_config,
                ),
            )

            params = dict(
                host=self._play_context.remote_addr,
                port=port,
                username=self._play_context.remote_user,
                password=self._play_context.password,
                key_filename=self.key_filename,
                hostkey_verify=self.get_option("host_key_checking"),
                look_for_keys=self.get_option("look_for_keys"),
                device_params=device_params,
                allow_agent=self._play_context.allow_agent,
                timeout=self.get_option("persistent_connect_timeout"),
                ssh_config=self._ssh_config,
            )
            # sock is only supported by ncclient >= 0.6.10, and will error if
            # included on older versions. We check the version in
            # _get_proxy_command, so if this returns a value, the version is
            # fine and we have something to send. Otherwise, don't even send
            # the option to support older versions of ncclient
            sock = self._get_proxy_command(port)
            if sock:
                params["sock"] = sock

            self._manager = manager.connect(**params)

            self._manager._timeout = self.get_option("persistent_command_timeout")
        except SSHUnknownHostError as exc:
            raise AnsibleConnectionFailure(to_native(exc))
        except AuthenticationError as exc:
            if str(exc).startswith("FileNotFoundError"):
                raise AnsibleError(
                    "Encountered FileNotFoundError in ncclient connect. Does {0} exist?".format(
                        self.key_filename
                    )
                )
            raise
        except ImportError:
            raise AnsibleError(
                "connection=netconf is not supported on {0}".format(self._network_os)
            )

        if not self._manager.connected:
            return 1, b"", b"not connected"

        self.queue_message("log", "ncclient manager object created successfully")

        self._connected = True

        super(Connection, self)._connect()

        return (
            0,
            to_bytes(self._manager.session_id, errors="surrogate_or_strict"),
            b"",
        )

    def close(self):
        if self._manager:
            self._manager.close_session()
        super(Connection, self).close()



# -*- coding: utf-8 -*-
# Copyright 2022 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

#############################################
#                WARNING                    #
#############################################
#
# This file is auto generated by the
# cli_rm_builder.
#
# Manually editing this file is not advised.
#
# To update the argspec make the desired changes
# in the module docstring and re-run
# cli_rm_builder.
#
#############################################

"""
The arg spec for the ios_bgp_global module
"""


class Bgp_globalArgs(object):  # pylint: disable=R0903
    """The arg spec for the ios_bgp_global module"""

    argument_spec = {
        "config": {
            "type": "dict",
            "options": {
                "as_number": {"type": "str"},
                "aggregate_addresses": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "netmask": {"type": "str"},
                        "advertise_map": {"type": "str"},
                        "as_confed_set": {"type": "bool"},
                        "as_set": {"type": "bool"},
                        "attribute_map": {"type": "str"},
                        "summary_only": {"type": "bool"},
                        "suppress_map": {"type": "str"},
                    },
                },
                "auto_summary": {"type": "bool"},
                "bgp": {
                    "type": "dict",
                    "options": {
                        "additional_paths": {
                            "type": "dict",
                            "options": {
                                "install": {"type": "bool"},
                                "receive": {"type": "bool"},
                                "select": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "best": {"type": "int"},
                                        "best_external": {"type": "bool"},
                                        "group_best": {"type": "bool"},
                                    },
                                },
                                "send": {"type": "bool"},
                            },
                        },
                        "advertise_best_external": {"type": "bool"},
                        "aggregate_timer": {"type": "int"},
                        "always_compare_med": {"type": "bool"},
                        "asnotation": {"type": "bool"},
                        "bestpath_options": {
                            "type": "dict",
                            "options": {
                                "aigp": {"type": "bool"},
                                "compare_routerid": {"type": "bool"},
                                "cost_community": {"type": "bool"},
                                "igp_metric": {"type": "bool"},
                                "med": {
                                    "type": "dict",
                                    "options": {
                                        "confed": {"type": "bool"},
                                        "missing_as_worst": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "client_to_client": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "all": {"type": "bool"},
                                "intra_cluster": {"type": "str"},
                            },
                        },
                        "cluster_id": {"type": "str"},
                        "confederation": {
                            "type": "dict",
                            "options": {"identifier": {"type": "str"}, "peers": {"type": "str"}},
                        },
                        "consistency_checker": {
                            "type": "dict",
                            "options": {
                                "auto_repair": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "interval": {"type": "int"},
                                    },
                                },
                                "error_message": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "interval": {"type": "int"},
                                    },
                                },
                            },
                        },
                        "dampening": {
                            "type": "dict",
                            "options": {
                                "penalty_half_time": {"type": "int"},
                                "reuse_route_val": {"type": "int"},
                                "suppress_route_val": {"type": "int"},
                                "max_suppress": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "default": {
                            "type": "dict",
                            "apply_defaults": True,
                            "options": {
                                "ipv4_unicast": {"type": "bool", "default": True},
                                "route_target": {
                                    "type": "dict",
                                    "apply_defaults": True,
                                    "options": {
                                        "filter": {"type": "bool", "default": True},
                                    },
                                },
                            },
                        },
                        "deterministic_med": {"type": "bool"},
                        "dmzlink_bw": {"type": "bool"},
                        "enforce_first_as": {"type": "bool"},
                        "enhanced_error": {"type": "bool"},
                        "fast_external_fallover": {"type": "bool"},
                        "graceful_restart": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "extended": {"type": "bool"},
                                "restart_time": {"type": "int"},
                                "stalepath_time": {"type": "int"},
                            },
                        },
                        "graceful_shutdown": {
                            "type": "dict",
                            "options": {
                                "neighbors": {
                                    "type": "dict",
                                    "options": {
                                        "time": {"type": "int"},
                                        "activate": {"type": "bool"},
                                    },
                                },
                                "vrfs": {
                                    "type": "dict",
                                    "options": {
                                        "time": {"type": "int"},
                                        "activate": {"type": "bool"},
                                    },
                                },
                                "community": {"type": "str"},
                                "local_preference": {"type": "int"},
                            },
                        },
                        "inject_maps": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "name": {"type": "str"},
                                "exist_map_name": {"type": "str"},
                                "copy_attributes": {"type": "bool"},
                            },
                        },
                        "listen": {
                            "type": "dict",
                            "options": {
                                "limit": {"type": "int"},
                                "range": {
                                    "type": "dict",
                                    "options": {
                                        "host_with_subnet": {"type": "str"},
                                        "peer_group": {"type": "str"},
                                    },
                                },
                            },
                        },
                        "log_neighbor_changes": {"type": "bool"},
                        "maxas_limit": {"type": "int"},
                        "maxcommunity_limit": {"type": "int"},
                        "maxextcommunity_limit": {"type": "int"},
                        "nexthop": {
                            "type": "dict",
                            "options": {
                                "route_map": {"type": "str"},
                                "trigger": {
                                    "type": "dict",
                                    "options": {
                                        "delay": {"type": "int"},
                                        "enable": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "nopeerup_delay_options": {
                            "type": "dict",
                            "options": {
                                "cold_boot": {"type": "int"},
                                "nsf_switchover": {"type": "int"},
                                "post_boot": {"type": "int"},
                                "user_initiated": {"type": "int"},
                            },
                        },
                        "recursion": {"type": "bool"},
                        "redistribute_internal": {"type": "bool"},
                        "refresh": {
                            "type": "dict",
                            "options": {
                                "max_eor_time": {"type": "int"},
                                "stalepath_time": {"type": "int"},
                            },
                        },
                        "regexp": {"type": "bool"},
                        "route_map": {"type": "bool"},
                        "router_id": {
                            "type": "dict",
                            "options": {
                                "address": {"type": "str"},
                                "interface": {"type": "str"},
                                "vrf": {"type": "bool"},
                            },
                        },
                        "scan_time": {"type": "int"},
                        "slow_peer": {
                            "type": "dict",
                            "options": {
                                "detection": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "threshold": {"type": "int"},
                                    },
                                },
                                "split_update_group": {
                                    "type": "dict",
                                    "options": {
                                        "dynamic": {"type": "bool"},
                                        "permanent": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "snmp": {"type": "bool"},
                        "sso": {"type": "bool"},
                        "soft_reconfig_backup": {"type": "bool"},
                        "suppress_inactive": {"type": "bool"},
                        "transport": {"type": "bool"},
                        "update_delay": {"type": "int"},
                        "update_group": {"type": "bool"},
                        "upgrade_cli": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "af_mode": {"type": "bool"}},
                        },
                    },
                },
                "bmp": {
                    "type": "dict",
                    "options": {
                        "buffer_size": {"type": "int"},
                        "initial_refresh": {
                            "type": "dict",
                            "options": {"delay": {"type": "int"}, "skip": {"type": "bool"}},
                        },
                        "server": {"type": "int"},
                        "server_options": {
                            "type": "dict",
                            "options": {
                                "activate": {"type": "bool"},
                                "address": {
                                    "type": "dict",
                                    "options": {"host": {"type": "str"}, "port": {"type": "int"}},
                                },
                            },
                        },
                    },
                },
                "default_information": {"type": "bool"},
                "default_metric": {"type": "int"},
                "distance": {
                    "type": "dict",
                    "options": {
                        "admin": {
                            "type": "dict",
                            "options": {
                                "distance": {"type": "int"},
                                "address": {"type": "str"},
                                "wildcard_bit": {"type": "str"},
                                "acl": {"type": "str"},
                            },
                        },
                        "bgp": {
                            "type": "dict",
                            "options": {
                                "routes_external": {"type": "int"},
                                "routes_internal": {"type": "int"},
                                "routes_local": {"type": "int"},
                            },
                        },
                        "mbgp": {
                            "type": "dict",
                            "options": {
                                "routes_external": {"type": "int"},
                                "routes_internal": {"type": "int"},
                                "routes_local": {"type": "int"},
                            },
                        },
                    },
                },
                "distributes": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "prefix": {"type": "str"},
                        "gateway": {"type": "str"},
                        "acl": {"type": "str"},
                        "in": {"type": "bool"},
                        "out": {"type": "bool"},
                        "interface": {"type": "str"},
                    },
                },
                "maximum_paths": {
                    "type": "dict",
                    "options": {
                        "paths": {"type": "int"},
                        "eibgp": {"type": "int"},
                        "ibgp": {"type": "int"},
                    },
                },
                "maximum_secondary_paths": {
                    "type": "dict",
                    "options": {
                        "paths": {"type": "int"},
                        "eibgp": {"type": "int"},
                        "ibgp": {"type": "int"},
                    },
                },
                "neighbors": {
                    "type": "list",
                    "elements": "dict",
                    "aliases": ["neighbor"],
                    "options": {
                        "neighbor_address": {"type": "str"},
                        "activate": {"type": "bool"},
                        "additional_paths": {
                            "type": "dict",
                            "options": {
                                "disable": {"type": "bool"},
                                "receive": {"type": "bool"},
                                "send": {"type": "bool"},
                            },
                        },
                        "advertise": {
                            "type": "dict",
                            "options": {
                                "additional_paths": {
                                    "type": "dict",
                                    "options": {
                                        "all": {"type": "bool"},
                                        "best": {"type": "int"},
                                        "group_best": {"type": "bool"},
                                    },
                                },
                                "best_external": {"type": "bool"},
                                "diverse_path": {
                                    "type": "dict",
                                    "options": {
                                        "backup": {"type": "bool"},
                                        "mpath": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "advertise_map": {
                            "type": "dict",
                            "options": {
                                "name": {"type": "str"},
                                "exist_map": {"type": "str"},
                                "non_exist_map": {"type": "str"},
                            },
                        },
                        "advertisement_interval": {"type": "int"},
                        "aigp": {
                            "type": "dict",
                            "options": {
                                "enable": {"type": "bool"},
                                "send": {
                                    "type": "dict",
                                    "options": {
                                        "cost_community": {
                                            "type": "dict",
                                            "options": {
                                                "id": {"type": "int"},
                                                "poi": {
                                                    "type": "dict",
                                                    "options": {
                                                        "igp_cost": {"type": "bool"},
                                                        "pre_bestpath": {"type": "bool"},
                                                        "transitive": {"type": "bool"},
                                                    },
                                                },
                                            },
                                        },
                                        "med": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "allow_policy": {"type": "bool"},
                        "allowas_in": {"type": "int"},
                        "as_override": {"type": "bool"},
                        "bmp_activate": {
                            "type": "dict",
                            "options": {"all": {"type": "bool"}, "server": {"type": "int"}},
                        },
                        "capability": {
                            "type": "dict",
                            "options": {
                                "both": {"type": "bool"},
                                "receive": {"type": "bool"},
                                "send": {"type": "bool"},
                            },
                        },
                        "cluster_id": {"type": "str"},
                        "default_originate": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "route_map": {"type": "str"}},
                        },
                        "description": {"type": "str"},
                        "disable_connected_check": {"type": "bool"},
                        "distribute_list": {
                            "type": "dict",
                            "options": {
                                "acl": {"type": "str"},
                                "in": {"type": "bool"},
                                "out": {"type": "bool"},
                            },
                        },
                        "dmzlink_bw": {"type": "bool"},
                        "ebgp_multihop": {
                            "type": "dict",
                            "options": {"enable": {"type": "bool"}, "hop_count": {"type": "int"}},
                        },
                        "fall_over": {
                            "type": "dict",
                            "options": {
                                "bfd": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "multi_hop": {"type": "bool"},
                                        "single_hop": {"type": "bool"},
                                    },
                                },
                                "route_map": {"type": "str"},
                            },
                        },
                        "filter_list": {
                            "type": "dict",
                            "options": {
                                "path_acl": {"type": "str"},
                                "in": {"type": "bool"},
                                "out": {"type": "bool"},
                            },
                        },
                        "ha_mode": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "disable": {"type": "bool"}},
                        },
                        "inherit": {"type": "str"},
                        "local_as": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "number": {"type": "int"},
                                "dual_as": {"type": "bool"},
                                "no_prepend": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "replace_as": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "log_neighbor_changes": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "disable": {"type": "bool"}},
                        },
                        "maximum_prefix": {
                            "type": "dict",
                            "options": {
                                "max_no": {"type": "int"},
                                "threshold_val": {"type": "int"},
                                "restart": {"type": "int"},
                                "warning_only": {"type": "bool"},
                            },
                        },
                        "next_hop_self": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "all": {"type": "bool"}},
                        },
                        "next_hop_unchanged": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "allpaths": {"type": "bool"}},
                        },
                        "password_options": {
                            "type": "dict",
                            "no_log": False,
                            "options": {
                                "encryption": {"type": "int"},
                                "pass_key": {"type": "str", "no_log": True},
                            },
                        },
                        "path_attribute": {
                            "type": "dict",
                            "options": {
                                "discard": {
                                    "type": "dict",
                                    "options": {
                                        "type": {"type": "int"},
                                        "range": {
                                            "type": "dict",
                                            "options": {
                                                "start": {"type": "int"},
                                                "end": {"type": "int"},
                                            },
                                        },
                                        "in": {"type": "bool"},
                                    },
                                },
                                "treat_as_withdraw": {
                                    "type": "dict",
                                    "options": {
                                        "type": {"type": "int"},
                                        "range": {
                                            "type": "dict",
                                            "options": {
                                                "start": {"type": "int"},
                                                "end": {"type": "int"},
                                            },
                                        },
                                        "in": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "peer_group": {"type": "str"},
                        "remote_as": {"type": "str"},
                        "remove_private_as": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "all": {"type": "bool"},
                                "replace_as": {"type": "bool"},
                            },
                        },
                        "route_maps": {
                            "type": "list",
                            "elements": "dict",
                            "options": {
                                "name": {"type": "str"},
                                "in": {"type": "bool"},
                                "out": {"type": "bool"},
                            },
                        },
                        "route_reflector_client": {"type": "bool"},
                        "route_server_client": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "context": {"type": "str"}},
                        },
                        "send_community": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "both": {"type": "bool"},
                                "extended": {"type": "bool"},
                                "standard": {"type": "bool"},
                            },
                        },
                        "send_label": {
                            "type": "dict",
                            "options": {"set": {"type": "bool"}, "explicit_null": {"type": "bool"}},
                        },
                        "shutdown": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "graceful": {"type": "int"},
                                "community": {"type": "int"},
                                "local_preference": {"type": "bool"},
                            },
                        },
                        "slow_peer": {
                            "type": "dict",
                            "options": {
                                "detection": {
                                    "type": "dict",
                                    "options": {
                                        "enable": {"type": "bool"},
                                        "disable": {"type": "bool"},
                                        "threshold": {"type": "int"},
                                    },
                                },
                                "split_update_group": {
                                    "type": "dict",
                                    "options": {
                                        "dynamic": {
                                            "type": "dict",
                                            "options": {
                                                "enable": {"type": "bool"},
                                                "disable": {"type": "bool"},
                                                "permanent": {"type": "bool"},
                                            },
                                        },
                                        "static": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "soft_reconfiguration": {"type": "bool"},
                        "timers": {
                            "type": "dict",
                            "options": {
                                "interval": {"type": "int"},
                                "holdtime": {"type": "int"},
                                "min_holdtime": {"type": "int"},
                            },
                        },
                        "translate_update": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "nlri": {
                                    "type": "dict",
                                    "options": {
                                        "multicast": {"type": "bool"},
                                        "unicast": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "transport": {
                            "type": "dict",
                            "options": {
                                "connection_mode": {
                                    "type": "dict",
                                    "options": {
                                        "active": {"type": "bool"},
                                        "passive": {"type": "bool"},
                                    },
                                },
                                "multi_session": {"type": "bool"},
                                "path_mtu_discovery": {
                                    "type": "dict",
                                    "options": {
                                        "set": {"type": "bool"},
                                        "disable": {"type": "bool"},
                                    },
                                },
                            },
                        },
                        "ttl_security": {"type": "int"},
                        "unsuppress_map": {"type": "str"},
                        "update_source": {"type": "str"},
                        "version": {"type": "int"},
                        "weight": {"type": "int"},
                    },
                },
                "networks": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "address": {"type": "str"},
                        "netmask": {"type": "str"},
                        "route_map": {"type": "str"},
                        "backdoor": {"type": "bool"},
                    },
                },
                "redistribute": {
                    "type": "list",
                    "elements": "dict",
                    "options": {
                        "application": {
                            "type": "dict",
                            "options": {
                                "name": {"type": "str"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "bgp": {
                            "type": "dict",
                            "options": {
                                "as_number": {"type": "str"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "connected": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "eigrp": {
                            "type": "dict",
                            "options": {
                                "as_number": {"type": "str"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "isis": {
                            "type": "dict",
                            "options": {
                                "area_tag": {"type": "str"},
                                "clns": {"type": "bool"},
                                "ip": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "iso_igrp": {
                            "type": "dict",
                            "options": {"area_tag": {"type": "str"}, "route_map": {"type": "str"}},
                        },
                        "lisp": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "mobile": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "odr": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "ospf": {
                            "type": "dict",
                            "options": {
                                "process_id": {"type": "int"},
                                "match": {
                                    "type": "dict",
                                    "options": {
                                        "external": {"type": "bool"},
                                        "internal": {"type": "bool"},
                                        "nssa_external": {"type": "bool"},
                                        "type_1": {"type": "bool"},
                                        "type_2": {"type": "bool"},
                                    },
                                },
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                                "vrf": {"type": "str"},
                            },
                        },
                        "ospfv3": {
                            "type": "dict",
                            "options": {
                                "process_id": {"type": "int"},
                                "match": {
                                    "type": "dict",
                                    "options": {
                                        "external": {"type": "bool"},
                                        "internal": {"type": "bool"},
                                        "nssa_external": {"type": "bool"},
                                        "type_1": {"type": "bool"},
                                        "type_2": {"type": "bool"},
                                    },
                                },
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "rip": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "static": {
                            "type": "dict",
                            "options": {
                                "set": {"type": "bool"},
                                "clns": {"type": "bool"},
                                "ip": {"type": "bool"},
                                "metric": {"type": "int"},
                                "route_map": {"type": "str"},
                            },
                        },
                        "vrf": {
                            "type": "dict",
                            "options": {"name": {"type": "str"}, "global": {"type": "bool"}},
                        },
                    },
                },
                "synchronization": {"type": "bool"},
                "table_map": {
                    "type": "dict",
                    "options": {"name": {"type": "str"}, "filter": {"type": "bool"}},
                },
                "template": {
                    "type": "dict",
                    "options": {"peer_policy": {"type": "str"}, "peer_session": {"type": "str"}},
                },
                "timers": {
                    "type": "dict",
                    "options": {
                        "keepalive": {"type": "int"},
                        "holdtime": {"type": "int"},
                        "min_holdtime": {"type": "int"},
                    },
                },
            },
        },
        "running_config": {"type": "str"},
        "state": {
            "choices": [
                "merged",
                "replaced",
                "overridden",
                "deleted",
                "purged",
                "gathered",
                "rendered",
                "parsed",
            ],
            "default": "merged",
            "type": "str",
        },
    }  # pylint: disable=C0301



# -*- coding: utf-8 -*-
# Copyright 2020 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function


__metaclass__ = type

"""
The Ospf_interfaces parser templates file. This contains
a list of parser definitions and associated functions that
facilitates both facts gathering and native command generation for
the given network resource.
"""

import re

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.rm_base.network_template import (
    NetworkTemplate,
)


class Ospf_interfacesTemplate(NetworkTemplate):
    def __init__(self, lines=None, module=None):
        super(Ospf_interfacesTemplate, self).__init__(lines=lines, tmplt=self, module=module)

    PARSERS = [
        {
            "name": "name",
            "getval": re.compile(
                r"""
                ^interface\s(?P<name>\S+)$""",
                re.VERBOSE,
            ),
            "setval": "interface {{ name }}",
            "result": {"{{ name }}": {"name": "{{ name }}", "address_family": {}}},
            "shared": True,
        },
        {
            "name": "process",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf
                (\s(?P<id>\d+))
                (\sarea\s(?P<area>\d+))?
                (\sarea\s(?P<area_ip>\s+))?
                (\s(?P<secondaries>secondaries))?
                (\sinstance\s(?P<instance>\d+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} {{ process.id|string }}"
            "{{ (' area ' + process.area_id|string ) if process.area_id is defined else '' }}"
            "{{ (' ' + secondaries) if process.secondaries is defined else '' }}"
            "{{ (' instance ' + process.instance) if process.instance is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "process": {
                                "id": "{{ id }}",
                                "area_id": "{{ area }}",
                                "secondaries": "{{ not not secondaries }}",
                                "instance_id": "{{ instance }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "adjacency",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf
                \s(?P<adjacency>adjacency\sstagger\sdisable)
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} adjacency stagger disable",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "adjacency": "{{ not not adjacency }}",
                        },
                    },
                },
            },
        },
        {
            "name": "authentication",
            "getval": re.compile(
                r"""
                \s+ip\sospf\sauthentication
                (\skey-chain\s(?P<key_chain>\S+))?
                (\s(?P<message_digest>message-digest))?
                (\s(?P<isnull>null))?
                $""",
                re.VERBOSE,
            ),
            "setval": "ip ospf authentication"
            "{{ (' key-chain ' + authentication.key_chain) if authentication.key_chain is defined else '' }}"
            "{{ (' ' + message-digest) if authentication.message_digest is defined else '' }}"
            "{{ (' ' + null) if authentication.null is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "authentication": {
                                "key_chain": "{{ key_chain }}",
                                "message_digest": "{{ not not message_digest }}",
                                "null": "{{ not not isnull }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "bfd",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sbfd
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} bfd",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "bfd": True,
                        },
                    },
                },
            },
        },
        {
            "name": "cost",
            "getval": re.compile(
                r"""
                \s+ip\sospf
                (\scost\s(?P<cost>\d+))
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} cost"
            " {{ cost.interface_cost|string }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "cost": {"interface_cost": "{{ cost }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "cost_ipv6_dynamic_cost",
            "getval": re.compile(
                r"""
                \s+ipv6\sospf\scost\s(?P<interface_cost>\d+)
                (\sdynamic)?
                (\sdefault\s(?P<default>\d+))?
                (\shysteresis)?
                (\spercent\s(?P<h_params_p>\d+))?
                (\sthreshold\s(?P<h_params_t>\d+))?
                (\sweight)?
                (\sL2-factor\s(?P<l2_factor>\d+))?
                (\slatency\s(?P<latency>\d+))?
                (\sresources\s(?P<resources>\d+))?
                (\sthroughput\s(?P<throughput>\d+))?
                (\s(?P<weight_oc>oc))?
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} cost"
            " {{ cost.interface_cost|string }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ipv6": {
                            "afi": "ipv6",
                            "cost": {
                                "interface_cost": "{{ interface_cost }}",
                                "dynamic_cost": {
                                    "default": "{{ default }}",
                                    "hysteresis": {
                                        "percent": "{{ h_params_p }}",
                                        "threshold": "{{ h_params_t  }}",
                                    },
                                    "weight": {
                                        "l2_factor": "{{ l2_factor  }}",
                                        "latency": "{{ latency  }}",
                                        "oc": "{{ not not weight_oc }}",
                                        "resources": "{{ resources  }}",
                                        "throughput": "{{ throughput }}",
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "database_filter",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sdatabase-filter\sall\sout
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} database-filter all out",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "database_filter": True,
                        },
                    },
                },
            },
        },
        {
            "name": "dead_interval",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sdead-interval
                (\s(?P<seconds>\d+))?
                (\sminimal\shello-multiplier\s(?P<minimal>\d+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} dead-interval {{ dead_interval.time|string }}"
            "{{ (' minimal hello-multiplier ' + dead_interval.minimal|string) if dead_interval.minimal is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "dead_interval": {"time": "{{ seconds }}", "minimal": "{{ minimal }}"},
                        },
                    },
                },
            },
        },
        {
            "name": "demand_circuit",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sdemand-circuit
                (\s(?P<ignore>ignore))?
                (\s(?P<disable>disable))?
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} demand-circuit"
            "{{ ' ignore' if demand_circuit.ignore is defined else '' }}"
            "{{ ' disable' if demand_circuit.disable is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "demand_circuit": {
                                "enable": True,
                                "ignore": "{{ not not ignore }}",
                                "disable": "{{ not not disable }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "flood_reduction",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sflood-reduction
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} flood-reduction",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "flood_reduction": True,
                        },
                    },
                },
            },
        },
        {
            "name": "hello_interval",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf(\shello-interval\s(?P<hello_interval>\d+))
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} hello-interval {{ hello_interval }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "hello_interval": "{{ hello_interval }}",
                        },
                    },
                },
            },
        },
        {
            "name": "lls",
            "getval": re.compile(
                r"""\s+ip\sospf\slls$""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} lls",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "lls": True,
                        },
                    },
                },
            },
        },
        {
            "name": "manet",
            "getval": re.compile(
                r"""
                \s+ipv6\sospf\smanet\speering
                (\scost\spercent\s(?P<cost_p>\d+))?
                (\scost\sthreshold\s(?P<cost_t>\d+))?
                (\slink-metrics\s(?P<link_metrics>\d+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "ipv6 ospf manet peering"
            "{{ ' cost' if manet.cost is defined else '' }}"
            "{{ (' percent ' + manet.cost.percent|string ) if manet.cost.percent is defined else '' }}"
            "{{ (' threshold ' + manet.cost.threshold|string ) if manet.cost.threshold is defined else '' }}"
            "{{ ' link-metrics' if manet.link_metrics is defined else '' }}"
            "{{ (' ' + manet.link_metrics.cost_threshold) if manet.link_metrics is defined and manet.link_metrics.cost_threshold is defined  else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ipv6": {
                            "afi": "ipv6",
                            "manet": {
                                "cost": {
                                    "percent": "{{ cost_p }}",
                                    "threshold": "{{ cost_t }}",
                                },
                                "link_metrics": {
                                    "set": "{{ True if link_metrics is not defined and link_metrics is defined  }}",
                                    "cost_threshold": "{{ link_metrics }}",
                                },
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "mtu_ignore",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\smtu-ignore
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} mtu-ignore",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "mtu_ignore": True,
                        },
                    },
                },
            },
        },
        {
            "name": "multi_area",
            "getval": re.compile(
                r"""
                \s+ip\sospf\smulti-area\s(?P<multi_area>\d+)
                (\scost\s(?P<cost>\d+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "ip ospf multi-area {{ multi_area.id|string }}"
            "{{ (' cost ' + multi_area.cost|string ) if multi_area.cost is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "multi_area": {
                                "id": "{{ multi_area }}",
                                "cost": "{{ cost }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "neighbor",
            "getval": re.compile(
                r"""
                \s+ipv6\sospf\sneighbor\s(?P<address>\S+)
                (\scost\s(?P<cost>\d+))?
                (\s(?P<database_filter>database-filter\sall\sout))?
                (\spoll-interval\s(?P<poll_interval>\d+))?
                (\spriority\s(?P<priority>\d+))?
                $""",
                re.VERBOSE,
            ),
            "setval": "ipv6 ospf neighbor {{ neighbor.address }}"
            "{{ (' cost ' + neighbor.cost|string ) if neighbor.cost is defined else '' }}"
            "{{ ' database-filter all out' if neighbor.database_filter is defined else '' }}"
            "{{ (' poll-interval ' + neighbor.poll_interval|string ) if neighbor.poll_interval is defined else '' }}"
            "{{ (' priority ' + neighbor.priority|string ) if neighbor.priority is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ipv6": {
                            "afi": "ipv6",
                            "neighbor": {
                                "address": "{{ address }}",
                                "cost": "{{ cost }}",
                                "database_filter": "{{ not not database_filter }}",
                                "poll_interval": "{{ poll_interval }}",
                                "priority": "{{ priority }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "network",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\snetwork
                (\s(?P<broadcast>broadcast))?
                (\s(?P<manet>manet))?
                (\s(?P<non_broadcast>non-broadcast))?
                (\s(?P<point_to_multipoint>point-to-multipoint))?
                (\s(?P<point_to_point>point-to-point))?
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} network"
            "{{ ' broadcast' if network.broadcast is defined else '' }}"
            "{{ ' manet' if network.manet is defined else '' }}"
            "{{ ' non-broadcast' if network.non_broadcast is defined else '' }}"
            "{{ ' point-to-multipoint' if network.point_to_multipoint is defined else '' }}"
            "{{ ' point-to-point' if network.point_to_point is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "network": {
                                "broadcast": "{{ not not broadcast }}",
                                "manet": "{{ not not manet }}",
                                "non_broadcast": "{{ not not non_broadcast }}",
                                "point_to_multipoint": "{{ not not point_to_multipoint }}",
                                "point_to_point": "{{ not not point_to_point }}",
                            },
                        },
                    },
                },
            },
        },
        {
            "name": "prefix_suppression",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sprefix-suppression
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} prefix-suppression",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "prefix_suppression": True,
                        },
                    },
                },
            },
        },
        {
            "name": "priority",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf
                (\spriority\s(?P<priority>\d+))
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} priority {{ priority }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "priority": "{{ priority }}",
                        },
                    },
                },
            },
        },
        {
            "name": "resync_timeout",
            "getval": re.compile(
                r"""
                \s+ip\sospf
                (\sresync-timeout\s(?P<resync_timeout>\d+))
                $""",
                re.VERBOSE,
            ),
            "setval": "ip ospf resync-timeout {{ resync_timeout|string }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "resync_timeout": "{{ resync_timeout }}",
                        },
                    },
                },
            },
        },
        {
            "name": "retransmit_interval",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf
                (\sretransmit-interval\s(?P<retransmit_interval>\d+))
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} retransmit-interval {{ retransmit_interval|string }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "retransmit_interval": "{{ retransmit_interval }}",
                        },
                    },
                },
            },
        },
        {
            "name": "shutdown",
            "getval": re.compile(
                r"""
                \s+(?P<afi>ip|ipv6)\sospf\sshutdown
                $""",
                re.VERBOSE,
            ),
            "setval": "{{ 'ip ospf' if afi == 'ipv4' else 'ipv6 ospf' }} shutdown",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "{{ afi }}": {
                            "afi": "{{ 'ipv4' if afi == 'ip' else 'ipv6' }}",
                            "shutdown": True,
                        },
                    },
                },
            },
        },
        {
            "name": "transmit_delay",
            "getval": re.compile(
                r"""
                \s+ipv6\sospf\stransmit-delay\s(?P<transmit_delay>\d+)
                $""",
                re.VERBOSE,
            ),
            "setval": "ipv6 ospf transmit-delay {{ transmit_delay|string }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ipv6": {
                            "afi": "ipv6",
                            "transmit_delay": "{{ transmit_delay }}",
                        },
                    },
                },
            },
        },
        {
            "name": "ttl_security",
            "getval": re.compile(
                r"""
                \s+ip\sospf\sttl-security\shops\s(?P<hops>\d+)
                *$""",
                re.VERBOSE,
            ),
            "setval": "ip ospf ttl-security"
            "{{ (' hops ' + ttl_security.hops|string) if ttl_security.hops is defined else '' }}",
            "result": {
                "{{ name }}": {
                    "address_family": {
                        "ip": {
                            "afi": "ipv4",
                            "ttl_security": {
                                "set": True,
                                "hops": "{{ hops }}",
                            },
                        },
                    },
                },
            },
        },
    ]



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (c) 2024, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

"""Ansible module to perform operations on SDA fabric transits in Cisco Catalyst Center."""
from __future__ import absolute_import, division, print_function

__metaclass__ = type
__author__ = ['Muthu Rakesh, Madhan Sankaranarayanan']

DOCUMENTATION = r"""
---
module: sda_fabric_transits_workflow_manager
short_description: Resource module for SDA fabric transits
description:
- Manage operations on SDA fabric transits.
- API to create transit networks.
- API to update transit networks.
- API to delete transit networks.
version_added: '6.18.0'
extends_documentation_fragment:
  - cisco.dnac.workflow_manager_params
author: Muthu Rakesh (@MUTHU-RAKESH-27)
        Madhan Sankaranarayanan (@madhansansel)
options:
  config_verify:
    description: Set to True to verify the Cisco Catalyst Center after applying the playbook config.
    type: bool
    default: False
  state:
    description: The state of Cisco Catalyst Center after module completion.
    type: str
    choices: [ merged, deleted ]
    default: merged
  config:
    description:
    - A list of SDA fabric transit configurations.
    - Each entry in the list represents a transit network configuration.
    type: list
    elements: dict
    required: true
    suboptions:
      sda_fabric_transits:
        description: SDA fabric transit configurations.
        type: list
        elements: dict
        suboptions:
          name:
            description:
            - The name of the SDA fabric transit.
            - It facilitates seamless communication between different network segments.
            - Required for the operations in the SDA fabric transits.
            type: str
          transit_type:
            description: Type of the fabric tranist.
                IP_BASED_TRANSIT - Responsible for managing IP routing and ensures data flow
                                   between various segments of the network.
                SDA_LISP_PUB_SUB_TRANSIT - Facilitates the decoupling of location and identity
                                   information for devices, enabling dynamic routing.
                SDA_LISP_BGP_TRANSIT - Integrates LISP with BGP to manage and optimize routing decisions
                                       between different network segments.
            default: IP_BASED_TRANSIT
            choices: [IP_BASED_TRANSIT, SDA_LISP_PUB_SUB_TRANSIT, SDA_LISP_BGP_TRANSIT]
            type: str
          ip_transit_settings:
            description:
            - The configuration settings for IP based transit.
            - Required when the type is set to IP_BASED_TRANSIT.
            - IP_BASED_TRANSIT cannot be updated.
            type: dict
            suboptions:
              routing_protocol_name:
                description: Defines the protocol for determining the best paths for data transmission between networks.
                type: str
                default: BGP
                choices: [BGP]
              autonomous_system_number:
                description:
                - Used by routing protocols like BGP to manage routing between different autonomous systems.
                - Autonomous System Number (ANS) should be from 1 to 4294967295.
                - The ASN should be unique for every IP-based transits.
                - Required when the transit_type is set to IP_BASED_TRANSIT.
                type: str
          sda_transit_settings:
            description:
            - The configuration settings for SDA-based transit.
            - Required when the transit_type is set to SDA_LISP_PUB_SUB_TRANSIT or SDA_LISP_BGP_TRANSIT.
            type: dict
            suboptions:
              is_multicast_over_transit_enabled:
                description:
                - Determines whether multicast traffic is permitted to traverse the transit network.
                - Enabling this option allows the distribution of data to multiple recipients across different network segments.
                - Available only when the transit type is set to SDA_LISP_PUB_SUB_TRANSIT.
                type: bool
              control_plane_network_device_ips:
                description:
                - Specifies the IP addresses of the network devices that form the control plane.
                - Required when the transit_type is set to either SDA_LISP_BGP_TRANSIT or SDA_LISP_PUB_SUB_TRANSIT.
                - Atleast one control plane network device is required.
                - A maximum of 2 control plane network devices are allowed when the transit_type is SDA_LISP_BGP_TRANSIT.
                - A maximum of 4 control plane network devices are allowed when the transit_type is SDA_LISP_PUB_SUB_TRANSIT.
                - SDA_LISP_PUB_SUB_TRANSIT supports only devices with IOS XE 17.6 or later.
                - The devices must be present in the Fabric site or zone.
                type: list
                elements: str

requirements:
- dnacentersdk >= 2.9.2
- python >= 3.9
notes:
  - SDK Method used are
    devices.Devices.get_device_list,
    sda.Sda.get_transit_networks,
    sda.Sda.add_transit_networks,
    sda.Sda.update_transit_networks,
    sda.Sda.delete_transit_network_by_id,
    task.Task.get_tasks_by_id,
    task.Task.get_task_details_by_id,

  - Paths used are
    get /dna/intent/api/v1/network-device,
    get /dna/intent/api/v1/sda/transitNetworks,
    post /dna/intent/api/v1/sda/transitNetworks,
    put /dna/intent/api/v1/sda/transitNetworks,
    delete /dna/intent/api/v1/sda/transitNetworks/${id},
    get /dna/intent/api/v1/tasks/${id}
    get /dna/intent/api/v1/tasks/${id}/detail

"""

EXAMPLES = r"""
- name: Create SDA fabric transit of transit_type IP_BASED_TRANSIT
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: IP_BASED_TRANSIT
      ip_transit_settings:
        routing_protocol_name: BGP
        autonomous_system_number: 1234

- name: Create SDA fabric transit of transit_type SDA_LISP_BGP_TRANSIT
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: SDA_LISP_BGP_TRANSIT
      sda_transit_settings:
        control_plane_network_device_ips:
        - string
        - string

- name: Create SDA fabric transit of transit_type SDA_LISP_PUB_SUB_TRANSIT
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: SDA_LISP_PUB_SUB_TRANSIT
      sda_transit_settings:
        is_multicast_over_transit_enabled: false
        control_plane_network_device_ips:
        - string
        - string
        - string
        - string

- name: Update SDA fabric transit of transit_type SDA_LISP_BGP_TRANSIT
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: SDA_LISP_BGP_TRANSIT
      sda_transit_settings:
        control_plane_network_device_ips:
        - string
        - string

- name: Update the multicast over transit
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: SDA_LISP_PUB_SUB_TRANSIT
      sda_transit_settings:
        is_multicast_over_transit_enabled: true

- name: Update the control plane network devices
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: merged
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit
      transit_type: SDA_LISP_PUB_SUB_TRANSIT
      sda_transit_settings:
        control_plane_network_device_ips:
        - string
        - string
        - string

- name: Delete SDA fabric transit
  cisco.dnac.sda_fabric_transits_workflow_manager:
    dnac_host: "{{dnac_host}}"
    dnac_username: "{{dnac_username}}"
    dnac_password: "{{dnac_password}}"
    dnac_verify: "{{dnac_verify}}"
    dnac_port: "{{dnac_port}}"
    dnac_version: "{{dnac_version}}"
    dnac_debug: "{{dnac_debug}}"
    dnac_log: True
    dnac_log_level: "{{ dnac_log_level }}"
    state: deleted
    config_verify: True
    config:
    - sda_fabric_transits:
      - name: sample_transit1
      - name: sample_transit2
"""

RETURN = r"""
# Case_1: Successful creation of SDA fabric transit
response_1:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

# Case_2: Successful updation of SDA fabric transit
response_2:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }

# Case_3: Successful deletion of SDA fabric transit
response_3:
  description: A dictionary or list with the response returned by the Cisco Catalyst Center Python SDK
  returned: always
  type: dict
  sample: >
    {
      "response": {
        "taskId": "string",
        "url": "string"
      },
      "version": "string"
    }
"""

import copy
from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.dnac.plugins.module_utils.dnac import (
    DnacBase,
    validate_list_of_dicts,
    get_dict_result,
    dnac_compare_equality,
)


class FabricTransit(DnacBase):
    """Class containing member attributes for sda_fabric_transits_workflow_manager module"""

    def __init__(self, module):
        super().__init__(module)
        self.response = [
            {"fabric_transits": {"response": {}, "msg": {}}}
        ]
        self.fabric_transits_obj_params = self.get_obj_params("fabricTransits")
        self.max_timeout = self.params.get('dnac_api_task_timeout')

    def validate_input(self):
        """
        Checks if the configuration parameters provided in the playbook
        meet the expected structure and data types,
        as defined in the 'temp_spec' dictionary.

        Parameters:
            self (object): The current object details.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Example:
            If the validation succeeds, 'self.status' will be 'success' and
            'self.validated_config' will contain the validated configuration.
            If it fails, 'self.status' will be 'failed', and
            'self.msg' will describe the validation issues.
        """

        if not self.config:
            self.msg = "config not available in playbook for validation."
            self.status = "success"
            return self

        # temp_spec is the specification for the expected structure of configuration parameters
        temp_spec = {
            "sda_fabric_transits": {
                "type": 'list',
                "elements": 'dict',
                "name": {"type": 'string'},
                "transit_type": {"type": 'string', "choices": ["IP_BASED_TRANSIT", "SDA_LISP_PUB_SUB_TRANSIT", "SDA_LISP_BGP_TRANSIT"]},
                "ip_transit_settings": {
                    "type": 'dict',
                    "routing_protocol_name": {"type": 'string', "choices": ["BGP"]},
                    "autonomous_system_number": {"type": 'integer'}
                },
                "sda_transit_settings": {
                    "type": 'dict',
                    "is_multicast_over_transit_enabled": {"type": 'bool'},
                    "control_plane_network_device_ips": {"type": 'list', "elements": 'string'}
                }
            }
        }

        # Validate playbook params against the specification (temp_spec)
        valid_temp, invalid_params = validate_list_of_dicts(self.config, temp_spec)
        if invalid_params:
            self.msg = (
                "Invalid parameters in playbook: {invalid_params}"
                .format(invalid_params="\n".join(invalid_params))
            )
            self.status = "failed"
            return self

        self.validated_config = valid_temp
        self.log("Successfully validated playbook config params: {valid_temp}"
                 .format(valid_temp=valid_temp), "INFO")
        self.msg = "Successfully validated input from the playbook."
        self.status = "success"
        return self

    def requires_update(self, have, want, obj_params):
        """
        Check if the config given requires update by comparing
        current information with the requested information.

        This method compares the current fabric transits from Cisco Catalyst Center
        with the user-provided details from the playbook,
        using a specified schema for comparison.

        Parameters:
            have (dict): Current information from the Cisco Catalyst Center
                          of SDA fabric transits.
            want (dict): Users provided information from the playbook
            obj_params (list of tuples) - A list of parameter mappings specifying which
                                          Cisco Catalyst Center parameters (dnac_param) correspond to
                                          the user-provided parameters (ansible_param).
        Returns:
            bool - True if any parameter specified in obj_params differs between
            current_obj and requested_obj, indicating that an update is required.
            False if all specified parameters are equal.
        Description:
            This function checks both the information provided by the user and
            the information available in the Cisco Catalyst Center.
            Based on the object_params the comparison will be taken place.
            If there is a difference in those information, it will return True.
            Else False.
        """

        current_obj = have
        requested_obj = want
        self.log("Current State (have): {current_obj}".format(current_obj=current_obj), "DEBUG")
        self.log("Desired State (want): {requested_obj}".format(requested_obj=requested_obj), "DEBUG")

        return any(not dnac_compare_equality(current_obj.get(dnac_param),
                                             requested_obj.get(ansible_param))
                   for (dnac_param, ansible_param) in obj_params)

    def get_obj_params(self, get_object):
        """
        Get the required comparison obj_params value

        Parameters:
            get_object (str): identifier for the required obj_params
        Returns:
            obj_params (list): obj_params value for comparison.
        Description:
            This function gets the object for the requires_update function.
            The obj_params will have the pattern to be compared.
        """

        try:
            if get_object == "fabricTransits":
                obj_params = [
                    ("isMulticastOverTransitEnabled", "isMulticastOverTransitEnabled"),
                    ("controlPlaneNetworkDeviceIds", "controlPlaneNetworkDeviceIds"),
                ]
            else:
                raise ValueError("Received an unexpected value for 'get_object': {object_name}"
                                 .format(object_name=get_object))
        except Exception as msg:
            self.log("Received exception: {msg}".format(msg=msg), "CRITICAL")

        return obj_params

    def get_device_details_by_ip(self, device_ip):
        """
        Get the network device details from the IP address.

        Parameters:
            device_ip (str): Playbook fabric transits details containing name,
            type, IP transits and SDA transits details.
        Returns:
            device_details (dict or None): The network device details with the given IP address.
        Description:
            This function calls the API to get the devices list based on the IP address.
            Returns the network device details.
        """

        response = self.dnac._exec(
            family="devices",
            function="get_device_list",
            op_modifies=True,
            params={"management_ip_address": device_ip}
        )
        if not isinstance(response, dict):
            self.log("Failed to retrieve the Authentication and Policy Server details - "
                     "Response is not a dictionary", "CRITICAL")
            return None

        device_details = response.get("response")
        if device_details:
            self.log("Successfully retrieved device details for IP " + str(device_ip), "INFO")
        else:
            self.log("No device details found for IP " + str(device_ip), "WARNING")

        return device_details

    def format_fabric_transit_params(self, fabric_transit_details):
        """
        Process the fabric transit parameters retrieved from the Cisco Catalyst Center

        Parameters:
            fabric_transit_details (dict): The fabric transit details from the Cisco Catalyst Center
        Returns:
            fabric_transit_info (dict): Processed fabric transit data in a format
            suitable for Cisco Catalyst Center configuration.
        Description:
            Formats the information with respect to the API payload.
        """

        fabric_transit_info = {
            "name": fabric_transit_details.get("name"),
            "type": fabric_transit_details.get("type")
        }
        ip_transit_settings = fabric_transit_details.get("ipTransitSettings")
        if ip_transit_settings:
            self.log("IP Transit settings found and processed.", "INFO")
            fabric_transit_info.update({"ipTransitSettings": ip_transit_settings})
            return fabric_transit_info

        sda_transit_settings = fabric_transit_details.get("sdaTransitSettings")
        fabric_transit_info.update({"sdaTransitSettings": sda_transit_settings})
        control_plane_devices = fabric_transit_info.get("sdaTransitSettings").get("controlPlaneNetworkDeviceIds")
        if control_plane_devices:
            sorted(control_plane_devices)

        return fabric_transit_info

    def fabric_transit_exists(self, name):
        """
        Check if the SDA fabric transit with the given name exists

        Parameters:
            name (str): The name of the fabric transit to check for existence.
        Returns:
            dict - A dictionary containing information about the
                   SDA fabric transit's existence:
                - 'exists' (bool): True if the fabric transit exists, False otherwise.
                - 'id' (str or None): The ID of the fabric transit if it exists or None if it doesn't.
                - 'details' (dict or None): Details of the fabric transit if it exists else None.
        Description:
            Sets the existance, details and the id of the fabric tranist as None.
            Calls the API 'get_transit_networks' until the transit is found or there are
            no more transits available in the Cisco Catalyst Center.
            If the transit is not found, the offset is incremented by 500, as the API returns
            a maximum of 500 entries at a time.
        """

        transit_info = {
            "exists": False,
            "details": None,
            "id": None
        }
        offset = 1
        while True:
            response = self.dnac._exec(
                family="sda",
                function="get_transit_networks",
                op_modifies=True,
                params={
                    "offset": offset
                }
            )
            if not isinstance(response, dict):
                self.msg = "Error in getting fabric transits - Response is not a dictionary"
                self.log(self.msg, "ERROR")
                self.status = "exited"
                return self.check_return_status()

            all_fabric_transit_details = response.get("response")
            if not all_fabric_transit_details:
                self.log("Fabric transit {name} does not exist.".format(name=name), "DEBUG")
                return transit_info

            fabric_transit_details = get_dict_result(all_fabric_transit_details, "name", name)
            if fabric_transit_details:
                self.log("Fabric transit found with name '{name}': {details}"
                         .format(name=name, details=fabric_transit_details), "INFO")
                transit_info.update({
                    "exists": True,
                    "id": fabric_transit_details.get("id"),
                    "details": self.format_fabric_transit_params(fabric_transit_details)
                })
                break

            offset += 500

        self.log("SDA fabric transit details: {details}".format(details=transit_info.get("details")), "DEBUG")
        self.log("SDA fabric transit id: {id}".format(id=transit_info.get("id")), "DEBUG")
        return transit_info

    def get_have_fabric_transits(self, fabric_transits):
        """
        Get the SDA fabric transits related information from Cisco
        Catalyst Center based on the provided playbook details.

        Parameters:
            fabric_transits (dict): Playbook details containing fabric transits details.
        Returns:
            self: The current object with updated current Fabric Transits information.
        Description:
            Sets the existance, details and the id of the fabric tranist as None.
            If the name is not available, return with a msg by setting the
            self.status as 'failed' and return self.
            Call a function which checks the existence of the fabric transit.
        """

        fabric_transits_details = []
        for item in fabric_transits:
            transit_info = {
                "exists": False,
                "details": None,
                "id": None
            }
            name = item.get("name")
            if not name:
                self.msg = "The required parameter 'name' in 'sda_fabric_transits' is missing."
                self.log(self.msg, "ERROR")
                self.status = "failed"
                return self

            transit_info = self.fabric_transit_exists(name)
            self.log("SDA transit transit exists for '{name}': {exists}"
                     .format(name=name, exists=transit_info.get("exists")), "DEBUG")
            self.log("SDA transit transit details for '{name}': {details}"
                     .format(name=name, details=transit_info.get("details")), "DEBUG")
            self.log("SDA transit transit Id for '{name}': {id}"
                     .format(name=name, id=transit_info.get("id")), "DEBUG")
            fabric_transits_details.append(transit_info)

        self.have.update({"fabric_transits": fabric_transits_details})
        self.msg = "Collecting the SDA fabric transits details from the Cisco Catalyst Center."
        self.status = "success"
        return self

    def get_have(self, config):
        """
        Get the SDA fabric transits related information from Cisco Catalyst Center

        Parameters:
            config (dict): Playbook details containing fabric transits details.
        Returns:
            self: The current object with updated fabric transits details.
        Description:
            Check the fabric_transits. If the name is not available, return with a msg by setting the
            self.status as 'failed' and return self.
            Call the 'get_have_fabric_transits' function to collect the information
            about the fabric transit in the Cisco Catalyst Center.
        """

        fabric_transits = config.get("sda_fabric_transits")
        if not fabric_transits:
            self.msg = "The parameter 'sda_fabric_transits' is missing under the 'config'."
            self.status = "failed"
            return self

        self.get_have_fabric_transits(fabric_transits).check_return_status()

        self.log("Current State (have): {current_state}".format(current_state=self.have), "INFO")
        self.msg = "Successfully retrieved the SDA fabric transits details from the Cisco Catalyst Center."
        self.status = "success"
        return self

    def handle_ip_transit_settings(self, item, fabric_transits_values, fabric_transit_index):
        """
        Handle the IP transit setting details.

        Parameters:
            item (dict): Playbook details containing fabric transits details.
            fabric_transits_values (dict): Partially stored fabric transit details.
            fabric_transit_index (int): Index which point the item in the fabric transit item in the config.
        Returns:
            fabric_transits_values (dict): Complete fabric transit details.
        Description:
            Check the basic cases and limits for all the parameters present in the SDA Fabric Transits.
        """

        fabric_transits_values.update({"ipTransitSettings": {}})
        ip_transit_settings = fabric_transits_values.get("ipTransitSettings")
        want_ip_transit_settings = item.get("ip_transit_settings")
        if not want_ip_transit_settings:
            self.msg = (
                "The 'ip_transit_settings' is mandatory when the 'transit_type' is set to 'IP_BASED_TRANSIT'."
            )
            self.status = "failed"
            return self.check_return_status()

        routing_protocol_name = item.get("ip_transit_settings").get("routing_protocol_name")
        routing_protocol_name_set = ("BGP")
        if not routing_protocol_name:
            routing_protocol_name = "BGP"

        if routing_protocol_name not in routing_protocol_name_set:
            self.msg = (
                "The 'routing_protocol_name' under 'ip_transit_settings' should be in the list: {routing_protocol_name_set}"
                .format(routing_protocol_name_set=routing_protocol_name_set)
            )
            self.status = "failed"
            return self.check_return_status()

        ip_transit_settings.update({"routingProtocolName": routing_protocol_name})
        have_fabric_transit_details = self.have.get("fabric_transits")[fabric_transit_index] \
                                               .get("details")
        have_ip_transit_settings = None
        if have_fabric_transit_details:
            have_ip_transit_settings = self.have.get("fabric_transits")[fabric_transit_index] \
                                                .get("details").get("ipTransitSettings")

        autonomous_system_number = item.get("ip_transit_settings").get("autonomous_system_number")
        if not autonomous_system_number:
            self.msg = "The required parameter 'autonomous_system_number' in 'ip_transit_settings' is missing."
            self.status = "failed"
            return self.check_return_status()

        try:
            autonomous_system_number_int = int(autonomous_system_number)
            if not 1 <= autonomous_system_number_int <= 4294967295:
                self.msg = "The 'autonomous_system_number' should be from 1 to 4294967295."
                self.status = "failed"
                return self.check_return_status()
        except ValueError:
            self.msg = "The 'autonomous_system_number' should contain only digits 0-9."
            self.status = "failed"
            return self.check_return_status()

        if have_ip_transit_settings:
            have_autonomous_system_number = self.have.get("fabric_transits")[fabric_transit_index].get("details") \
                                                     .get("ipTransitSettings").get("autonomousSystemNumber")
            if not autonomous_system_number:
                autonomous_system_number = have_autonomous_system_number
            elif have_autonomous_system_number != str(autonomous_system_number):
                self.msg = "The 'autonomous_system_number' in 'ip_transit_settings' cannot be updated."
                self.status = "failed"
                return self.check_return_status()

        ip_transit_settings.update({"autonomousSystemNumber": str(autonomous_system_number)})

        return fabric_transits_values

    def remove_duplicate_ips(self, control_plane_ips):
        """
        Remove the duplicates from the given list.

        Parameters:
            control_plane_ips (list): List of elements which may contain duplicates.
        Returns:
            final_control_plane_ips (list): List of elements with out duplicates.
        Description:
            Return empty list if the list is empty or it is a NoneType. Check whether any
            duplicates is present in the list or not. If yes, remove them and return the list.
        """

        self.log(
            "The list of control plane ips before removing the duplicates {list_of_ips}"
            .format(list_of_ips=control_plane_ips), "DEBUG"
        )
        final_control_plane_ips = []

        # No need to proceed when there is no elements in the list
        if not control_plane_ips:
            self.log("Received an empty or None list. Returning an empty list.", "DEBUG")
            return final_control_plane_ips

        control_plane_ips = sorted(control_plane_ips)
        self.log(
            "Control plane IPs sorted: {0}".format(control_plane_ips),
            "DEBUG"
        )
        length_control_plane_ips = len(control_plane_ips)

        # No need to check for the duplicates when there is only one element in the list
        if length_control_plane_ips == 1:
            self.log("Only one IP found, no duplicates to remove.", "DEBUG")
            return control_plane_ips

        final_control_plane_ips.append(control_plane_ips[0])
        for i in range(1, length_control_plane_ips):
            if control_plane_ips[i] != control_plane_ips[i - 1]:
                final_control_plane_ips.append(control_plane_ips[i])

        self.log(
            "The list of control plane IPs after removing the duplicates '{list_of_ips}'"
            .format(list_of_ips=final_control_plane_ips), "DEBUG"
        )

        return final_control_plane_ips

    def handle_sda_transit_settings(self, item, fabric_transits_values, transit_type, fabric_transit_index):
        """
        Handle the SDA transit settings details.

        Parameters:
            item (dict): Playbook details containing fabric transits details.
            fabric_transits_values (dict): Partially stored fabric transit details.
            transit_type (str): Type of the fabric transit.
            fabric_transit_index (int): Index which point the item in the fabric transit item in the config.
        Returns:
            fabric_transits_values (dict): Complete fabric transit details.
        Description:
            Check the basic cases and limits for all the parameters present in the SDA Fabric Transits.
        """

        fabric_transits_values.update({"sdaTransitSettings": {}})
        have_sda_transit_settings = None
        have_fabric_details = self.have.get("fabric_transits")[fabric_transit_index].get("details")
        if have_fabric_details:
            have_sda_transit_settings = have_fabric_details.get("sdaTransitSettings")

        sda_transit_settings = fabric_transits_values.get("sdaTransitSettings")
        want_sda_transit_settings = item.get("sda_transit_settings")
        if want_sda_transit_settings:
            if transit_type == "SDA_LISP_PUB_SUB_TRANSIT":
                is_multicast_over_transit_enabled = item.get("sda_transit_settings").get("is_multicast_over_transit_enabled")
                if is_multicast_over_transit_enabled is not None:
                    sda_transit_settings.update({"isMulticastOverTransitEnabled": is_multicast_over_transit_enabled})
                elif have_sda_transit_settings:
                    sda_transit_settings.update({"isMulticastOverTransitEnabled": have_sda_transit_settings.get("isMulticastOverTransitEnabled")})
                else:
                    sda_transit_settings.update({"isMulticastOverTransitEnabled": False})

            control_plane_network_device_ips = self.remove_duplicate_ips(want_sda_transit_settings.get("control_plane_network_device_ips"))
            if have_sda_transit_settings and not control_plane_network_device_ips:
                sda_transit_settings.update({"controlPlaneNetworkDeviceIds": sorted(have_sda_transit_settings.get("controlPlaneNetworkDeviceIds"))})
            elif control_plane_network_device_ips:
                length_of_control_plane_network_device_ips = len(control_plane_network_device_ips)
                if transit_type == "SDA_LISP_BGP_TRANSIT":
                    max_ips = 2
                else:
                    max_ips = 4
                if transit_type == "SDA_LISP_BGP_TRANSIT" and length_of_control_plane_network_device_ips > max_ips:
                    self.msg = "The 'control_plane_network_device_ips' should have a maximum of 2 IPs when the 'transit_type' is 'SDA_LISP_BGP_TRANSIT'."
                    self.status = "failed"
                    return self.check_return_status()
                elif transit_type == "SDA_LISP_PUB_SUB_TRANSIT" and length_of_control_plane_network_device_ips > max_ips:
                    self.msg = "The 'control_plane_network_device_ips' should have a maximum of 4 IPs when the 'transit_type' is 'SDA_LISP_PUB_SUB_TRANSIT'."
                    self.status = "failed"
                    return self.check_return_status()

                control_plane_network_device_ids = []
                for network_device_ip in control_plane_network_device_ips:
                    network_device_details = self.get_device_details_by_ip(network_device_ip)
                    if not network_device_details:
                        self.msg = (
                            "There is no network device in the Cisco Catalyst Center with the IP '{device_ip}'."
                            .format(device_ip=network_device_ip)
                        )
                        self.status = "failed"
                        return self.check_return_status()

                    network_device_id = network_device_details[0].get("instanceUuid")
                    control_plane_network_device_ids.append(network_device_id)
                sda_transit_settings.update({"controlPlaneNetworkDeviceIds": sorted(control_plane_network_device_ids)})
            else:
                self.msg = "The parameter 'control_plane_network_device_ips' in 'sda_transit_settings' should not be empty."
                self.status = "failed"
                return self.check_return_status()

        elif have_sda_transit_settings and not want_sda_transit_settings:
            fabric_transits_values.update({"sdaTransitSettings": have_sda_transit_settings})
            sorted(fabric_transits_values.get("sdaTransitSettings").get("controlPlaneNetworkDeviceIds"))
        else:
            self.msg = "The parameter 'sda_transit_settings' should not be empty."
            self.status = "failed"
            return self.check_return_status()

        return fabric_transits_values

    def get_want_fabric_transits(self, fabric_transits):
        """
        Get all the SDA fabric transits information from playbook
        Set the status and the msg before returning from the API
        Check the return value of the API with check_return_status()

        Parameters:
            fabric_transits (dict): Playbook fabric transits details containing name,
            transit_type, IP transits and SDA transits details.
        Returns:
            self: The current object with updated desired Fabric Transits information.
        Description:
            Do all the validation which is done in the GUI and format the payload
            in such a way that we can pass it to the API.
            If any of the validation fails, return with a msg by setting the
            self.status as 'failed' and return self.
        """

        want_fabric_transits = []
        fabric_transit_index = 0
        for item in fabric_transits:
            fabric_transits_values = {}
            name = item.get("name")
            if not name:
                self.msg = "The required parameter 'name' in 'sda_fabric_transits' is missing."
                self.status = "failed"
                return self

            if len(name) > 32:
                self.msg = "The length of the 'name' in 'sda_fabric_transits' should be less or equal to 32."
                self.status = "failed"
                return self

            fabric_transits_values.update({"name": name})
            state = self.params.get("state")
            if state == "deleted":
                want_fabric_transits.append(fabric_transits_values)
                fabric_transit_index += 1
                continue

            transit_type = item.get("transit_type")
            transit_type_set = ("IP_BASED_TRANSIT", "SDA_LISP_PUB_SUB_TRANSIT", "SDA_LISP_BGP_TRANSIT")
            have_type = None
            fabric_transit_details = self.have.get("fabric_transits")[fabric_transit_index].get("details")
            if fabric_transit_details:
                have_type = fabric_transit_details.get("type")

            if not transit_type:
                if have_type:
                    transit_type = have_type
                else:
                    transit_type = "IP_BASED_TRANSIT"

            if transit_type not in transit_type_set:
                self.msg = (
                    "The 'transit_type' under 'sda_fabric_transits' should be in the list: {transit_types}"
                    .format(transit_types=transit_type_set)
                )
                self.status = "failed"
                return self

            if have_type and transit_type != have_type:
                self.msg = (
                    "The parameter 'transit_type' under the SDA fabric transit '{name}' cannot be updated."
                    .format(name=name)
                )
                self.status = "failed"
                return self

            fabric_transits_values.update({"type": transit_type})
            if transit_type == "IP_BASED_TRANSIT":
                fabric_transits_values = copy.deepcopy(self.handle_ip_transit_settings(item,
                                                                                       fabric_transits_values,
                                                                                       fabric_transit_index))
            else:
                fabric_transits_values = copy.deepcopy(self.handle_sda_transit_settings(item,
                                                                                        fabric_transits_values,
                                                                                        transit_type,
                                                                                        fabric_transit_index))

            want_fabric_transits.append(fabric_transits_values)
            fabric_transit_index += 1

        self.log("SDA fabric transits playbook details: {requested_state}"
                 .format(requested_state=want_fabric_transits), "DEBUG")
        self.want.update({"fabric_transits": want_fabric_transits})
        self.msg = "Collecting the SDA fabric transits details from the playbook"
        self.status = "success"
        return self

    def get_want(self, config):
        """
        Get the SDA fabric transits related information from playbook.

        Parameters:
            config (list of dict): Playbook details
        Returns:
            self: The current object with updated fabric transits details.
        Description:
            Check the fabric_transits. If the name is not available, return with a msg by setting the
            self.status as 'failed' and return self.
            Call the 'get_want_fabric_transits' function to collect the information
            about the fabric transit passed by the user.
        """

        fabric_transits = config.get("sda_fabric_transits")
        if not fabric_transits:
            self.msg = "The parameter 'sda_fabric_transits' is missing under the 'config'."
            self.status = "failed"
            return self

        self.get_want_fabric_transits(fabric_transits).check_return_status()

        self.log("Desired State (want): {requested_state}".format(requested_state=self.want), "INFO")
        self.msg = "Successfully retrieved details from the playbook"
        self.status = "success"
        return self

    def update_fabric_transits(self, fabric_transits):
        """
        Create/Update fabric transit in Cisco Catalyst Center with fields provided in playbook.

        Parameters:
            fabric_transits (list of dict): SDA fabric transit playbook details.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            Check if the fabric transit is present in the Cisco Catalys Center or not.
            If not, call the API 'add_transit_networks' to create the  fabric transit. Else, check for the update.
            Call the requires_update, if the transit requires does not an update, return self after updating the msg.
            Or Call the API 'update_transit_networks' to update the transit.
            Update the result and return self.
        """

        fabric_transit_index = -1
        for item in fabric_transits:
            fabric_transit_index += 1
            name = item.get("name")
            result_fabric_transit = self.response[0].get("fabric_transits")
            have_fabric_transit = self.have.get("fabric_transits")[fabric_transit_index]
            want_fabric_transit = self.want.get("fabric_transits")[fabric_transit_index]
            self.log("Current SDA fabric transit '{name}' details in Catalyst Center: {current_details}"
                     .format(name=name, current_details=have_fabric_transit.get("details")), "DEBUG")
            self.log("Desired SDA fabric transit '{name}' details for Catalyst Center: {requested_details}"
                     .format(name=name, requested_details=want_fabric_transit), "DEBUG")

            # Check transit exists, if not create and continue
            if not have_fabric_transit.get("exists"):
                self.log("Desired fabric transit '{name}' details (want): {requested_state}"
                         .format(name=name, requested_state=want_fabric_transit), "DEBUG")
                payload = {"payload": [want_fabric_transit]}
                task_name = "add_transit_networks"
                task_id = self.get_taskid_post_api_call("sda", task_name, payload)
                if not task_id:
                    self.msg = (
                        "Unable to retrive the task_id for the task '{task_name}'."
                        .format(task_name=task_name)
                    )
                    self.set_operation_result("failed", False, self.msg, "ERROR")
                    return self

                success_msg = (
                    "Successfully created the SDA Transit with the details '{transit_details}'."
                    .format(transit_details=want_fabric_transit)
                )
                self.get_task_status_from_tasks_by_id(task_id, task_name, success_msg).check_return_status()
                self.log("Successfully created SDA fabric transit '{name}'.".format(name=name), "INFO")
                result_fabric_transit.get("response").update({
                    name: want_fabric_transit
                })
                result_fabric_transit.get("msg").update({
                    name: "SDA fabric transit created successfully"
                })
                continue

            # Check update is required

            if want_fabric_transit.get("type") == "IP_BASED_TRANSIT" or \
                    not self.requires_update(have_fabric_transit.get("details").get("sdaTransitSettings"),
                                             want_fabric_transit.get("sdaTransitSettings"),
                                             self.fabric_transits_obj_params):
                self.log("SDA fabric transit '{name}' doesn't require a update".format(name=name), "INFO")
                result_fabric_transit.get("msg").update({
                    name: "SDA fabric transit doesn't require an update."
                })
                continue

            self.log("Updating SDA fabric transit '{name}'.".format(name=name), "DEBUG")

            # Tranist Exists
            self.log("Current SDA fabric transit '{name}' details in Catalyst Center: {current_state}"
                     .format(name=name, current_state=have_fabric_transit), "DEBUG")
            self.log("Desired SDA fabric transit '{name}' details: {requested_state}"
                     .format(name=name, requested_state=want_fabric_transit), "DEBUG")
            want_fabric_transit.update({"id": have_fabric_transit.get("id")})
            payload = {"payload": [want_fabric_transit]}
            task_name = "update_transit_networks"
            task_id = self.get_taskid_post_api_call("sda", task_name, payload)
            if not task_id:
                self.msg = (
                    "Unable to retrive the task_id for the task '{task_name}'."
                    .format(task_name=task_name)
                )
                self.set_operation_result("failed", False, self.msg, "ERROR")
                return self

            success_msg = (
                "Successfully updated the SDA Transit with the details '{transit_details}'."
                .format(transit_details=want_fabric_transit)
            )
            self.get_task_status_from_tasks_by_id(task_id, task_name, success_msg).check_return_status()
            self.log("SDA fabric transit '{name}' updated successfully.".format(name=name), "INFO")
            result_fabric_transit.get("response").update({
                name: want_fabric_transit
            })
            result_fabric_transit.get("msg").update({
                name: "SDA fabric transit updated successfully."
            })

        self.log("Updated the SDA fabric transits successfully", "INFO")
        return self

    def get_diff_merged(self, config):
        """
        Create or Update the SDA fabric transits in Cisco Catalyst Center based on the playbook details.

        Parameters:
            config (list of dict): Playbook details containing SDA fabric transits information.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            If the 'sda_fabric_transits' is available in the playbook, call the function 'update_fabric_transits'.
            Else return self.
        """

        fabric_transits = config.get("sda_fabric_transits")
        if fabric_transits is not None:
            self.update_fabric_transits(fabric_transits)

        self.result.update({
            "response": self.response
        })
        return self

    def delete_fabric_transits(self, fabric_transits):
        """
        Delete fabric transit in Cisco Catalyst Center with fields provided in playbook.

        Parameters:
            fabric_transits (list of dict): SDA fabric transit playbook details.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            Check if the fabric transit is present in the Cisco Catalys Center or not.
            If not, set the result and return self.
            Or Call the API 'delete_transit_network_by_id' to delete the transit.
            Update the result and return self.
        """

        fabric_transit_index = -1
        for item in fabric_transits:
            fabric_transit_index += 1
            name = item.get("name")
            have_fabric_transit = self.have.get("fabric_transits")[fabric_transit_index]
            result_fabric_transit = self.response[0].get("fabric_transits")

            if not have_fabric_transit.get("exists"):
                result_fabric_transit.get("msg").update({name: "SDA fabric transit not found."})
                self.log("SDA fabric transit '{name}' not found".format(name=name), "INFO")
                continue

            self.log("SDA fabric transit scheduled for deletion with the name '{name}'.".format(name=name), "INFO")
            transit_id = have_fabric_transit.get("id")
            self.log("SDA fabric transit '{name}' id: {id}".format(name=name, id=transit_id), "DEBUG")
            payload = {"id": transit_id}
            task_name = "delete_transit_network_by_id"
            task_id = self.get_taskid_post_api_call("sda", task_name, payload)
            if not task_id:
                self.msg = (
                    "Unable to retrive the task_id for the task '{task_name}'."
                    .format(task_name=task_name)
                )
                self.set_operation_result("failed", False, self.msg, "ERROR")
                return self

            success_msg = (
                "Successfully deleted the SDA Transit with id '{id}'."
                .format(id=transit_id)
            )
            self.get_task_status_from_tasks_by_id(task_id, task_name, success_msg).check_return_status()
            result_fabric_transit.get("response").update({name: {}})
            result_fabric_transit.get("response").get(name).update({
                "Task Id": task_id
            })
            result_fabric_transit.get("msg").update({
                name: "SDA fabric transit deleted successfully"
            })

        self.msg = "SDA fabric transit(s) deleted successfully."
        self.status = "success"
        return self

    def get_diff_deleted(self, config):
        """
        Delete the SDA fabric transits in Cisco Catalyst Center based on the playbook details.

        Parameters:
            config (list of dict): Playbook details containing SDA fabric transits information.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            If the 'sda_fabric_transits' is available in the playbook, call the function 'delete_fabric_transits'.
            Else return self.
        """

        fabric_transits = config.get("sda_fabric_transits")
        if fabric_transits is not None:
            self.delete_fabric_transits(fabric_transits)

        self.result.update({
            "response": self.response
        })
        return self

    def verify_diff_merged(self, config):
        """
        Validating the Cisco Catalyst Center configuration with the playbook details
        when state is merged (Create/Update).

        Parameters:
            config (dict): Playbook details containing fabric transit configuration.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            If the 'sda_fabric_transits' is available in the playbook, collect the fabric transit information
            from the Cisco Catalyst Center.
            Check if there is any differece between the config passed by the user and the informaion
            collected by the user.
            If there is any differece set the msg by settings the self.status as 'failed' and return self.
        """

        self.get_have(config)
        self.log("Current State (have): {current_state}".format(current_state=self.have), "INFO")
        self.log("Requested State (want): {requested_state}".format(requested_state=self.want), "INFO")
        if config.get("sda_fabric_transits") is not None:
            want_fabric_transits = self.want.get("fabric_transits")
            have_fabric_transits = self.have.get("fabric_transits")
            self.log("Current State of SDA fabric transits (have): {current_details}"
                     .format(current_details=have_fabric_transits), "DEBUG")
            self.log("Desired State of SDA fabric transits (want): {requested_details}"
                     .format(requested_details=want_fabric_transits), "DEBUG")
            fabric_transit_index = 0
            for item in want_fabric_transits:
                fabric_transit_details = have_fabric_transits[fabric_transit_index].get("details")
                if not fabric_transit_details:
                    self.msg = (
                        "The SDA fabric transit config is not created with the config: {config}"
                        .format(config=item)
                    )
                    self.status = "failed"
                    return self

                if fabric_transit_details.get("type") == "IP_BASED_TRANSIT":
                    fabric_transit_index += 1
                    continue

                self.log("Current SDA Transit Settings: " + str(fabric_transit_details.get("sdaTransitSettings")))
                self.log("Desired SDA Transit Settings: " + str(item.get("sdaTransitSettings")))
                if self.requires_update(fabric_transit_details.get("sdaTransitSettings"),
                                        item.get("sdaTransitSettings"),
                                        self.fabric_transits_obj_params):
                    self.msg = "The SDA fabric transit config is not applied to the Cisco Catalyst Center."
                    self.status = "failed"
                    return self

                fabric_transit_index += 1

            self.log("Successfully validated SDA fabric transit(s).", "INFO")
            self.response[0].get("fabric_transits").update({"Validation": "Success"})

        self.result.update({
            "response": self.response
        })
        self.msg = "Successfully validated the SDA fabric transit(s)."
        self.status = "success"
        return self

    def verify_diff_deleted(self, config):
        """
        Validating the Cisco Catalyst Center configuration with the playbook details
        when state is deleted (delete).

        Parameters:
            config (dict): Playbook details containing fabric transit configuration.
        Returns:
            self (object): The current object with updated desired Fabric Transits information.
        Description:
            If the 'sda_fabric_transits' is available in the playbook, collect the fabric transit information
            from the Cisco Catalyst Center.
            Check if there is any transit with the name given in the playbook, set the msg by settings the
            self.status as 'failed' and return self.
        """

        self.get_have(config)
        self.log("Current State (have): {current_state}".format(current_state=self.have), "INFO")
        self.log("Desired State (want): {requested_state}".format(requested_state=self.want), "INFO")
        fabric_transits = config.get("sda_fabric_transits")
        if fabric_transits is not None:
            fabric_transit_index = 0
            fabric_transit_details = self.have.get("fabric_transits")
            for item in fabric_transit_details:
                fabric_transit_exists = item.get("exists")
                name = config.get("sda_fabric_transits")[fabric_transit_index].get("name")
                if fabric_transit_exists:
                    self.msg = (
                        "The SDA fabric transit config '{name}' is still present in "
                        "the Cisco Catalyst Center.".format(name=name)
                    )
                    self.status = "failed"
                    return self

                self.log("Successfully validated absence of transit '{name}'.".format(name=name), "INFO")
                fabric_transit_index += 1
            self.response[0].get("fabric_transits").update({"Validation": "Success"})

        self.result.update({
            "response": self.response
        })
        self.msg = "Successfully validated the absence of SDA fabric transit(s)."
        self.status = "success"
        return self

    def reset_values(self):
        """
        Reset all neccessary attributes to default values

        Parameters:
            self (object): The current object details.
        Returns:
            None
        """

        self.have.clear()
        self.want.clear()
        return


def main():
    """main entry point for module execution"""

    # Define the specification for module arguments
    element_spec = {
        "dnac_host": {"type": 'str', "required": True},
        "dnac_port": {"type": 'str', "default": '443'},
        "dnac_username": {"type": 'str', "default": 'admin', "aliases": ['user']},
        "dnac_password": {"type": 'str', "no_log": True},
        "dnac_verify": {"type": 'bool', "default": 'True'},
        "dnac_version": {"type": 'str', "default": '2.2.3.3'},
        "dnac_debug": {"type": 'bool', "default": False},
        "dnac_log": {"type": 'bool', "default": False},
        "dnac_log_level": {"type": 'str', "default": 'WARNING'},
        "dnac_log_file_path": {"type": 'str', "default": 'dnac.log'},
        "dnac_log_append": {"type": 'bool', "default": True},
        "config_verify": {"type": 'bool', "default": False},
        "dnac_api_task_timeout": {"type": 'int', "default": 1200},
        "dnac_task_poll_interval": {"type": 'int', "default": 2},
        "config": {"type": 'list', "required": True, "elements": 'dict'},
        "state": {"default": 'merged', "choices": ['merged', 'deleted']},
        "validate_response_schema": {"type": 'bool', "default": True},
    }

    # Create an AnsibleModule object with argument specifications
    module = AnsibleModule(argument_spec=element_spec, supports_check_mode=False)
    ccc_sda_transit = FabricTransit(module)
    if ccc_sda_transit.compare_dnac_versions(ccc_sda_transit.get_ccc_version(), "2.3.7.6") < 0:
        ccc_sda_transit.msg = (
            "The specified version '{0}' does not support the SDA fabric devices feature. Supported versions start from '2.3.7.6' onwards. "
            "Version '2.3.7.6' introduces APIs for creating, updating and deleting the IP and SDA Transits."
            .format(ccc_sda_transit.get_ccc_version())
        )
        ccc_sda_transit.status = "failed"
        ccc_sda_transit.check_return_status()

    state = ccc_sda_transit.params.get("state")
    config_verify = ccc_sda_transit.params.get("config_verify")
    if state not in ccc_sda_transit.supported_states:
        ccc_sda_transit.status = "invalid"
        ccc_sda_transit.msg = "State '{state}' is invalid".format(state=state)
        ccc_sda_transit.check_return_status()

    ccc_sda_transit.validate_input().check_return_status()

    for config in ccc_sda_transit.config:
        ccc_sda_transit.reset_values()
        ccc_sda_transit.get_have(config).check_return_status()
        if state != "deleted":
            ccc_sda_transit.get_want(config).check_return_status()
        ccc_sda_transit.get_diff_state_apply[state](config).check_return_status()
        if config_verify:
            ccc_sda_transit.verify_diff_state_apply[state](config).check_return_status()

    module.exit_json(**ccc_sda_transit.result)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# This file is part of Ansible
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

try:
    import botocore
except ImportError:
    pass  # caught by AnsibleAWSModule

from ansible_collections.amazon.aws.plugins.module_utils.botocore import is_boto3_error_code
from ansible_collections.amazon.aws.plugins.module_utils.tagging import ansible_dict_to_boto3_tag_list
from ansible_collections.amazon.aws.plugins.module_utils.tagging import boto3_tag_list_to_ansible_dict
from ansible_collections.amazon.aws.plugins.module_utils.tagging import compare_aws_tags


def manage_tags(module, client, resource_type, resource_id, new_tags, purge_tags):
    if new_tags is None:
        return False

    old_tags = get_tags(module, client, resource_type, resource_id)
    tags_to_set, tags_to_delete = compare_aws_tags(old_tags, new_tags, purge_tags=purge_tags)

    change_params = dict()
    if tags_to_set:
        change_params["AddTags"] = ansible_dict_to_boto3_tag_list(tags_to_set)
    if tags_to_delete:
        change_params["RemoveTagKeys"] = tags_to_delete

    if not change_params:
        return False

    if module.check_mode:
        return True

    try:
        client.change_tags_for_resource(ResourceType=resource_type, ResourceId=resource_id, **change_params)
    except (botocore.exceptions.BotoCoreError, botocore.exceptions.ClientError) as e:
        module.fail_json_aws(
            e,
            msg=f"Failed to update tags on {resource_type}",
            resource_id=resource_id,
            change_params=change_params,
        )
    return True


def get_tags(module, client, resource_type, resource_id):
    try:
        tagset = client.list_tags_for_resource(
            ResourceType=resource_type,
            ResourceId=resource_id,
        )
    except is_boto3_error_code("NoSuchHealthCheck"):
        return {}
    except is_boto3_error_code("NoSuchHostedZone"):  # pylint: disable=duplicate-except
        return {}
    except (
        botocore.exceptions.BotoCoreError,
        botocore.exceptions.ClientError,
    ) as e:  # pylint: disable=duplicate-except
        module.fail_json_aws(e, msg=f"Failed to fetch tags on {resource_type}", resource_id=resource_id)

    tags = boto3_tag_list_to_ansible_dict(tagset["ResourceTagSet"]["Tags"])
    return tags



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2015, Matt Davis <mdavis_ansible@rolpdog.com>
# Copyright: (c) 2017, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
---
module: win_updates
short_description: Download and install Windows updates
description:
    - Searches, downloads, and installs Windows updates synchronously by automating the Windows Update client.
options:
    accept_list:
        description:
        - A list of update titles or KB numbers that can be used to specify
          which updates are to be searched or installed.
        - If an available update does not match one of the entries, then it
          is skipped and not installed.
        - Each entry can either be the KB article or Update title as a regex
          according to the PowerShell regex rules.
        - The accept list is only validated on updates that were found based on
          I(category_names). It will not force the module to install an update
          if it was not in the category specified.
        type: list
        elements: str
    category_names:
        description:
        - A scalar or list of categories to install updates from. To get the list
          of categories, run the module with C(state=searched). The category must
          be the full category string, but is case insensitive.
        - Some possible categories are Application, Connectors, Critical Updates,
          Definition Updates, Developer Kits, Feature Packs, Guidance, Security
          Updates, Service Packs, Tools, Update Rollups, Updates, and Upgrades.
        - Since C(v1.7.0) the value C(*) will match all categories.
        type: list
        elements: str
        default: [ CriticalUpdates, SecurityUpdates, UpdateRollups ]
    skip_optional:
        description:
        - Skip optional updates where the update has BrowseOnly set by Microsoft.
        - Microsoft documents show that BrowseOnly means that the update
          should not be installed automatically and appear as optional updates.
        type: bool
        default: no
        version_added: 1.8.0
    reboot:
        description:
        - Ansible will automatically reboot the remote host if it is required
          and continue to install updates after the reboot.
        - This can be used instead of using a M(ansible.windows.win_reboot) task after this one
          and ensures all updates for that category is installed in one go.
        - Async does not work when C(reboot=true).
        type: bool
        default: no
    reboot_timeout:
        description:
        - The time in seconds to wait until the host is back online from a
          reboot.
        - This is only used if C(reboot=true) and a reboot is required.
        default: 1200
        type: int
    server_selection:
        description:
        - Defines the Windows Update source catalog.
        - C(default) Use the default search source. For many systems default is
          set to the Microsoft Windows Update catalog. Systems participating in
          Windows Server Update Services (WSUS) or similar corporate update server environments may
          default to those managed update sources instead of the Windows Update
          catalog.
        - C(managed_server) Use a managed server catalog. For environments
          utilizing Windows Server Update Services (WSUS) or similar corporate update servers, this
          option selects the defined corporate update source.
        - C(windows_update) Use the Microsoft Windows Update catalog.
        type: str
        choices: [ default, managed_server, windows_update ]
        default: default
    state:
        description:
        - Controls whether found updates are downloaded or installed or listed
        - This module also supports Ansible check mode, which has the same effect as setting state=searched
        type: str
        choices: [ installed, searched, downloaded ]
        default: installed
    log_path:
        description:
        - If set, C(win_updates) will append update progress to the specified file. The directory must already exist.
        type: path
    reject_list:
        description:
        - A list of update titles or KB numbers that can be used to specify
          which updates are to be excluded from installation.
        - If an available update does match one of the entries, then it is
          skipped and not installed.
        - Each entry can either be the KB article or Update title as a regex
          according to the PowerShell regex rules.
        type: list
        elements: str
    _operation:
        description:
        - Internal use only.
        type: str
        choices:
        - start
        - cancel
        - poll
        default: start
    _operation_options:
        description:
        - Internal use only.
        type: dict
notes:
- M(ansible.windows.win_updates) must be run by a user with membership in the local Administrators group.
- M(ansible.windows.win_updates) will use the default update service configured for the machine (Windows Update, Microsoft Update, WSUS, etc).
- By default M(ansible.windows.win_updates) does not manage reboots, but will signal when a
  reboot is required with the I(reboot_required) return value.
  I(reboot) can be used to reboot the host if required in the one task.
- M(ansible.windows.win_updates) can take a significant amount of time to complete (hours, in some cases).
  Performance depends on many factors, including OS version, number of updates, system load, and update server load.
- Beware that just after M(ansible.windows.win_updates) reboots the system, the Windows system may not have settled yet
  and some base services could be in limbo. This can result in unexpected behavior.
  Check the examples for ways to mitigate this.
- More information about PowerShell and how it handles RegEx strings can be
  found at U(https://technet.microsoft.com/en-us/library/2007.11.powershell.aspx).
- The current module doesn't support Systems Center Configuration Manager (SCCM).
  See U(https://github.com/ansible-collections/ansible.windows/issues/194)
- By default the C(ansible.builtin.ssh) connection plugin is configured to have
  no server timeout. As Windows Updates can restart the network adapter it is
  recommended to set C(-o ServerAliveInterval=30) and disable control master
  in I(ansible_ssh_args) to ensure the client can handle a network reset.
  See the examples showing one way this can be set.
- By default the module will start a background process using the Task
  Scheduler on Windows. If the Task Scheduler is unavailable, unreliable, or
  does not work, run the task with become.
seealso:
- module: chocolatey.chocolatey.win_chocolatey
- module: ansible.windows.win_feature
- module: community.windows.win_hotfix
- module: ansible.windows.win_package
author:
- Matt Davis (@nitzmahone)
"""

EXAMPLES = r"""
- name: Install all updates and reboot as many times as needed
  ansible.windows.win_updates:
    category_names: '*'
    reboot: true

- name: Set a server alive interval during update stage for the ssh connection plugin
  ansible.windows.win_updates:
    category_names: '*'
    reboot: true
  vars:
    # This can be set in a few ways, see the ssh connection plugin for more
    # information. ControlMaster should be disabled to ensure the new timeout
    # value is applied for this connection instead of through the cached
    # connection.
    ansible_ssh_args: -o ControlMaster=no -o ServerAliveInterval=30

- name: Install all security, critical, and rollup updates without a scheduled task
  ansible.windows.win_updates:
    category_names:
      - SecurityUpdates
      - CriticalUpdates
      - UpdateRollups
  become: true
  become_method: runas
  become_user: SYSTEM

- name: Search-only, return list of found updates (if any), log to C:\ansible_wu.txt
  ansible.windows.win_updates:
    category_names: SecurityUpdates
    state: searched
    log_path: C:\ansible_wu.txt

- name: Install all security updates with automatic reboots
  ansible.windows.win_updates:
    category_names:
      - SecurityUpdates
    reboot: true

- name: Install only particular updates based on the KB numbers
  ansible.windows.win_updates:
    category_names:
      - SecurityUpdates
    accept_list:
      - KB4056892
      - KB4073117

- name: Exclude updates based on the update title
  ansible.windows.win_updates:
    category_names:
      - SecurityUpdates
      - CriticalUpdates
    reject_list:
      - Windows Malicious Software Removal Tool for Windows
      - \d{4}-\d{2} Cumulative Update for Windows Server 2016

# Optionally, you can increase the reboot_timeout to survive long updates during reboot
- name: Ensure we wait long enough for the updates to be applied during reboot
  ansible.windows.win_updates:
    reboot: true
    reboot_timeout: 3600

# Search and download Windows updates
- name: Search and download Windows updates without installing them
  ansible.windows.win_updates:
    state: downloaded
"""

RETURN = r"""
reboot_required:
    description: True when the target server requires a reboot to complete updates (no further updates can be installed until after a reboot).
    returned: success
    type: bool
    sample: true

rebooted:
    description:
    - Set to C(true) when the target Windows host has been rebooted by C(win_updates).
    returned: success
    type: bool
    sample: false
    version_added: 1.14.0

updates:
    description:
    - Updates that were found/installed.
    - The key for each update is the C(id) of the update.
    returned: success
    type: dict
    sample:
    contains:
        title:
            description: Display name.
            returned: always
            type: str
            sample: "Security Update for Windows Server 2012 R2 (KB3004365)"
        kb:
            description: A list of KB article IDs that apply to the update.
            returned: always
            type: list
            elements: str
            sample: [ '3004365' ]
        id:
            description: Internal Windows Update GUID.
            returned: always
            type: str
            sample: "fb95c1c8-de23-4089-ae29-fd3351d55421"
        downloaded:
            description: Was the update downloaded.
            returned: always
            type: bool
            sample: true
            version_added: 1.7.0
        installed:
            description: Was the update successfully installed.
            returned: always
            type: bool
            sample: true
        categories:
            description: A list of category strings for this update.
            returned: always
            type: list
            elements: str
            sample: [ 'Critical Updates', 'Windows Server 2012 R2' ]
        failure_hresult_code:
            description: The HRESULT code from a failed update.
            returned: on install or download failure
            type: bool
            sample: 2147942402
        failure_msg:
            description: The error message with more details on the failure.
            returned: on install or download failure and not running with async
            type: str
            sample: Operation did not complete because there is no logged-on interactive user (WU_E_NO_INTERACTIVE_USER 0x80240020)
            version_added: 1.7.0

filtered_updates:
    description: Updates that were found but were filtered based on
      I(blacklist), I(whitelist) or I(category_names). The return value is in
      the same form as I(updates), along with I(filtered_reason).
    returned: success
    type: dict
    sample: see the updates return value
    contains:
        filtered_reason:
            description:
            - The reason why this update was filtered.
            - This value has been deprecated since C(1.7.0), use C(filtered_reasons) which contain a list of all the
              reasons why the update is filtered.
            returned: always
            type: str
            sample: 'skip_hidden'
        filtered_reasons:
            description:
            - A list of reasons why the update has been filtered.
            - Can be C(accept_list), C(reject_list), C(hidden), C(category_names), or C(skip_optional).
            type: list
            elements: str
            sample:
            - category_names
            - accept_list
            version_added: 1.7.0

found_update_count:
    description: The number of updates found needing to be applied.
    returned: success
    type: int
    sample: 3
installed_update_count:
    description: The number of updates successfully installed or downloaded.
    returned: success
    type: int
    sample: 2
failed_update_count:
    description: The number of updates that failed to install.
    returned: always
    type: int
    sample: 0
"""
